"""Module parallelization of viperleed.calc.lib.

Defines functionality useful for running tasks in multiple processes.

Part of this functionality used to be in lib.leedbase (originally
2019-06-13). Moved here especially to prevent cyclic imports.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-22'
__license__ = 'GPLv3+'

import logging
import multiprocessing
from pathlib import Path
import time

import psutil

from viperleed.calc.files import parameters


logger = logging.getLogger(__name__)


def monitoredPool(rp, poolsize, function, tasks, update_from=Path()):
    """
    The 'function' and 'tasks' arguments are passed on to a multiprocessing
    pool of size 'poolsize' with apply_async. While waiting for the pool to
    finish, the PARAMETERS file is read every second to check whether there is
    a STOP command. If so, the pool is terminated.

    Parameters
    ----------
    rp : Rparams object
        Needed for the parameter update
    poolsize : int
        passed on to multiprocessing.Pool
    function : function
        passed on to multiprocessing.Pool.apply_async
    tasks : list of arguments
        treated like the arguments of pool.map, i.e. each element is passed on
        in a seperate call of 'function' via multiprocessing.Pool.apply_async
    update_from : pathlike
        directory from which PARAMETERS should be read for updates

    Returns
    -------
    None

    """

    def kill_pool(p):
        """Kill the subprocesses, then terminate the pool."""
        for proc in p._pool:
            parent = psutil.Process(proc.pid)
            for child in parent.children(recursive=True):
                child.kill()
        p.terminate()

    def checkPoolResult(r):
        nonlocal pool
        nonlocal killed
        if r != "":
            kill_pool(pool)
            killed = True
        return r

    pool = multiprocessing.Pool(poolsize)
    results = []
    killed = False
    for task in tasks:
        r = pool.apply_async(function, (task,), callback=checkPoolResult)
        results.append(r)
    pool.close()
    try:
        while not all(r.ready() for r in results):
            if killed:
                break
            # See if user wants to STOP
            parameters.update(rp, update_from=update_from)
            if rp.STOP:
                kill_pool(pool)
                logger.info("Stopped by STOP parameter.")
                return
            time.sleep(1)
    except KeyboardInterrupt:
        logger.warning("Stopped by keyboard interrupt.")
        kill_pool(pool)
        raise
    pool.join()
    error = False
    for r in results:
        try:
            v = r.get(timeout=1)
            if v:
                logger.error(v)
                error = True
        except (TimeoutError, multiprocessing.context.TimeoutError):
            logger.error("Failed to get result from execution of {}"
                         .format(function.__name__))
            error = True
    if error:
        raise RuntimeError("Error in parallel execution of {}"
                           .format(function.__name__))



"""Module search_cull of viperleed.calc.classes.rparams.special.

Defines the SearchCullType and SearchCull classes. They are convenience
classes for handling user input of the SEARCH_CULL parameter.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-12-14'
__license__ = 'GPLv3+'

from dataclasses import InitVar, dataclass, field
from enum import Enum  # Unfortunately StrEnum was introduced in py3.11
from numbers import Real

from ..defaults import NO_VALUE
from .base import SpecialParameter


class SearchCullType(Enum):
    """Enumeration of the implemented culling strategies."""

    GENETIC = 'genetic'
    CLONE = 'clone'
    RANDOM = 'random'
    UNSET = NO_VALUE

    @property
    def is_clone(self):
        """Return whether this is a GENETIC cull type."""
        return self is SearchCullType.CLONE

    @property
    def is_genetic(self):
        """Return whether this is a GENETIC cull type."""
        return self is SearchCullType.GENETIC

    @property
    def is_random(self):
        """Return whether this is a GENETIC cull type."""
        return self is SearchCullType.RANDOM


@dataclass(repr=False)
class SearchCull(SpecialParameter, param='SEARCH_CULL'):
    """A container for information used when culling a search population.

    Attributes
    ----------
    value : int or float
        When a integer, corresponds to the exact number of
        individuals to be culled from a search population.
        When a float, it is a fraction of the population.
        Use the nr_individuals(population) method to always
        retrieve the absolute number of individuals to be
        culled.
    type_ : SearchCullType
        Which culling scheme should be applied. Currently
        supported schemes are:
            CLONE (duplicate some individuals)
            GENETIC (combine traits from pairs of individuals)
            RANDOM (pick random traits)
    """

    _value: InitVar[Real]
    _type: SearchCullType = field(default=NO_VALUE, repr=False)
    _int: int = field(init=False, default=NO_VALUE, repr=False)
    _fraction: float = field(init=False, default= NO_VALUE, repr=False)

    def __post_init__(self, _value):
        """Assign integer or fractional values for this SearchCull."""
        # Most of the functionality in here used to
        # be part of the ParameterInterpreter class
        try:
            cull_float = float(_value)
        except (TypeError, ValueError):
            raise TypeError('SearchCull value should be numeric') from None

        self._assign_int_or_fraction(cull_float)

        # Assign the type_ @property to make sure we correctly attempt
        # to convert the initialization argument to a SearchCullType
        self.type_ = self._type

    def _assign_int_or_fraction(self, cull_float):
        """Assign self._int or self._fraction from cull_float."""
        # See if we have an integer-valued cull
        cull_int = int(cull_float)
        if cull_float >= 1 and abs(cull_float - cull_int) < 1e-6:
            self._int = cull_int
        elif cull_float >= 1:
            raise ValueError('Values greater than one must be integers')
        elif cull_float >= 0:
            self._fraction = cull_float
        else:
            raise ValueError('SearchCull value must be non-negative')

    def __bool__(self):
        """Return whether this SearchCull would cause any culling."""
        return self.value > 0

    def __repr__(self):
        """Return a representation string for this SearchCull."""
        return f'{self.__class__.__name__}({self.value}, type_={self.type_})'

    @property
    def has_type(self):
        """Return whether this SearchCull has a type_ defined."""
        return self.type_ is not SearchCullType.UNSET

    @property
    def is_fractional(self):
        """Return whether this SearchCull is a fraction of the population."""
        return self._fraction is not NO_VALUE

    @property
    def type_(self):
        """Return the type of this SearchCull."""
        return self._type

    @type_.setter
    def type_(self, new_type):
        """Set the type of this SearchCull."""
        try:
            self._type = SearchCullType(new_type)
        except ValueError:
            raise ValueError(f'Unknown SearchCull type {new_type!r}') from None

    @property
    def value(self):
        """Return the (integer or fractional) value for this SearchCull."""
        return self._fraction if self.is_fractional else self._int

    @classmethod
    def from_value(cls, value):
        """Return a SearchCull from a Sequence value."""
        return cls(*value)

    def nr_individuals(self, population):
        """Return the number of individuals to be culled from population."""
        if not isinstance(population, int) or population < 0:
            raise ValueError(f'Invalid population={population!r}. '
                             'Should be a non-negative integer')
        ncull = (round(self.value * population) if self.is_fractional
                 else self.value)
        if ncull > population:
            raise ValueError(f'{self} too large for {population} individuals')
        return ncull




import numpy as np



# read in delta amplitudes into numpy array


# get intensities on the fly for parameter combination




def read_delta(rp, sl, delta_file):
    # reading of delta files by Tobias

    return
    #

# implement MPI eventually

# what is the grid?
# deal with grid, etc.

# Interpolate the deltas
# either numpy interp or scipy.interpolate. If n-dimensional, only linear interpolation is possible


#

# Optimization loop

#@jit
def rfactor_form_search_pars(real_pars, params_realation, offsets):
    """Callable R-factor function for the structure search which is passed to
     optimization algorithms found in scipy.minimize. Returns the R-factor of a given set
     of experimental and theoretical beams (with deltas)."""



    #expand_params
    #call rfactor
    R = 1

    return R




# what are the independent parameters; how do I get them; how do I get the others from there?

#symref is geometric relation, but what gives the index realation ?


def create_search_params_array(rp):
    #
    geo_idp_params = []
    geo_all_params = []
    geo_dep_params = []

    for parameter in rp.searchpars:
        # for now we only work with geometrical deltas
        if parameter.mode in ['occ', 'dom']:
            continue
        geo_all_params.append(parameter)
        if parameter.linkedTo is None:
            # independent parameter
            geo_idp_params.append(parameter)
        else:
            geo_dep_params.append(parameter)

    #sanity checks
    assert(len(geo_idp_params) == rp.indyPars)

    n_pars = len(geo_idp_params) + len(geo_dep_params)

    assert(n_pars == len(rp.search_atlist))

    # make array of parameters
    real_params = np.zeros([len(geo_idp_params)])
    # bounds as required by scipy.optimize
    real_params_bounds = []
    for idp_id, idp_par in enumerate(geo_idp_params):
        real_params[idp_id] = idp_par.center - 1 # -1 here because center is a Fortran index
        real_params_bounds.append((0, idp_par.steps -1)) # -1 here because center is a Fortran index

    params_relation = np.zeros([len(geo_idp_params), n_pars], dtype="int32")

    #create the correspondence matrix between dependent and independent atoms
    for par_id, param in enumerate(geo_all_params):
        for idp_id, idp_param in enumerate(geo_idp_params):
            if param.linkedTo == idp_param or param == idp_param:
                params_relation[idp_id, par_id] = 1

    # sum must be same as number of parameters in total
    assert(np.sum(params_relation) == len(geo_all_params))

    return real_params, real_params_bounds, params_relation, geo_idp_params, geo_all_params

def get_nc_surf(rp, sl, geo_all_params):
    nc_surf = np.full([len(geo_all_params)], False, dtype="bool") # flag if atom is considered to be at the surface or not
    surface_atoms = sl.getSurfaceAtoms(rp)
    for id, par in enumerate(geo_all_params):
        if par.atom in surface_atoms:
            nc_surf[id] = True

    return nc_surf

#@jit
def expand_params(real_params, params_relation, offsets):
    """extends the set of real pars into the set of all pars"""
    all_pars = np.dot(real_params, params_relation) + offsets

    return all_pars





"""Section Error calculation."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2021-03-18'
__license__ = 'GPLv3+'

import copy
import logging
import os
from pathlib import Path

from viperleed.calc.classes.r_error import R_Error
from viperleed.calc.classes.searchpar import SearchPar
from viperleed.calc.files import ioerrorcalc
from viperleed.calc.files.displacements import readDISPLACEMENTS_block
from viperleed.calc.sections.deltas import deltas as section_deltas
from viperleed.calc.sections.rfactor import rfactor as section_rfactor
from viperleed.calc.sections.superpos import superpos as section_superpos


logger = logging.getLogger(__name__)


def errorcalc(sl, rp):
    """Runs 1D error calculations based on DISPLACEMENTS."""

    if rp.domainParams:
        # TODO!! requested by Tilman!!
        # !!! Should be straightforward. Just take the beams from other domains
        # as constant (from refcalc), vary for the one domain as always
        logger.error("Error calculation not implemented for multiple domains.")
        return

    if rp.best_v0r is None:
        logger.info("Error calculation called without a stored inner "
                    "potential shift. Running R-factor calculation "
                    "from refcalc-fd.out to determine inner potential shift.")
        section_rfactor(sl, rp, index=11)
        logger.info("Finished R-factor pre-run, now starting error "
                    "calculation.\n")

    errors = []
    if rp.search_index > 0 and rp.search_index >= len(rp.disp_blocks):
        # running after one or more searches; read the last DISPLACEMENTS block
        rp.search_index -= 1
    seg_info = {"geo": "geometrical", "vib": "vibrational",
                "occ": "occupation"}
    for mode in "geo", "vib", "occ":
        sl.restoreOriState()  # reset positions, store any changes as offsets
        # read DISPLACEMENTS block - ONLY geo OR vib
        deltas_required = readDISPLACEMENTS_block(
            rp,
            sl,
            rp.disp_blocks[rp.search_index],
            only_mode=mode
            )
        rp.disp_block_read = True  # to prevent deltas segment from re-reading
        if not deltas_required:
            continue
        logger.info("\nStarting error calculations for "
                    + seg_info[mode] + " displacements.")
        # run delta calculations
        section_deltas(sl, rp)
        sl.deltas_initialized = True
        if rp.STOP:     # since this may stop the deltas, also check here
            return
        rp.generateSearchPars(sl)
        # group atoms under variation by linking
        atom_groups = [[at] for at in rp.search_atlist
                       if [sp for sp in rp.searchpars
                           if sp.atom == at and sp.steps > 1]]
        i = 0
        while i < len(atom_groups):
            found = None
            for at in atom_groups[i]:
                sps = [sp for sp in rp.searchpars if
                       sp.atom == at and sp.mode == mode and sp.el != "vac"
                       and ((isinstance(sp.linkedTo, SearchPar)
                             and sp.linkedTo.atom not in atom_groups[i]) or
                            (isinstance(sp.restrictTo, SearchPar)
                             and sp.restrictTo.atom not in atom_groups[i]))]
                if not sps:
                    continue
                if isinstance(sps[0].linkedTo, SearchPar):
                    found = [ag for ag in atom_groups
                             if sps[0].linkedTo.atom in ag][0]
                else:
                    found = [ag for ag in atom_groups
                             if sps[0].restrictTo.atom in ag][0]
                if found:
                    break
            if found:  # merge the two groups
                found.extend(atom_groups[i])
                atom_groups.pop(i)
            else:
                i += 1
        # run superpos and rfactor:
        for ag in atom_groups:
            # rp.search_atlist = ag  # limit variations to only that group
            only_vary = [sp for sp in rp.searchpars if sp.atom in ag]
            logger.info("\nNow calculating " + seg_info[mode] + " errors for "
                        "atom group: " + ", ".join(str(at) for at in ag))
            logger.info("Running superpos...")
            section_superpos(sl, rp, for_error=True, only_vary=only_vary)
            if rp.halt >= rp.HALTING:
                return
            if os.path.isfile("ROUTSHORT"):
                os.remove("ROUTSHORT")
            logger.info("Starting R-factor calculation...")
            rfaclist = section_rfactor(sl, rp, index=12, for_error=True,
                                       only_vary=only_vary)
            logger.info("Finished with " + seg_info[mode] + " errors for "
                        "atom group: " + ", ".join(str(at) for at in ag))
            error_disp_label = ag[0].displist[0].disp_labels[mode]
            error_lin_disps = ag[0].displist[0].disp_lin_steps[mode]
            errors.append(R_Error(rp.R_FACTOR_TYPE,
                                  ag, mode, rfaclist,
                                  error_disp_label,
                                  error_lin_disps,
                                  v0i=rp.V0_IMAG,
                                  energy_range=rp.total_energy_range()))
            if rp.halt >= rp.HALTING:
                return
    if len(errors) == 0:
        logger.info("Error calculation: Returning with no output.")
        return

    # Inform user that statistical error estimates are only available
    # for Pendry R-factor.
    if rp.R_FACTOR_TYPE != 1:
        logger.info("Estimates for statistical uncertainties "
                    "of parameters are only available for the Pendry "
                    "R-factor.")
    else:
        # Write var(R) to log as info.
        var_r_info = ioerrorcalc.extract_var_r(errors)
        if any(var_r_info.values()):
            var_str = []
            for mode, var_r in var_r_info.items():
                if not var_r:
                    continue
                var_str.append(
                    f"{mode}: {ioerrorcalc.format_col_content(var_r)}"
                    )
            var_str = ", ".join(var_str)
            logger.info(f"Found values for var(R): {var_str}")
        else:
            logger.info("Could not estimate var(R) for any error mode.")

    save_path = rp.workdir

    # Errors_summary.csv and Errors.zip
    (summary_csv_content,
     individual_files) = ioerrorcalc.generate_errors_csv(errors)
    ioerrorcalc.write_errors_archive(
        individual_files,
        archive_path=save_path,
        compression_level=rp.ZIP_COMPRESSION_LEVEL,
        archive_fname="Errors.zip"
        )
    ioerrorcalc.write_errors_summary_csv(summary_csv_content,
                                         summary_path=save_path,
                                         summary_fname="Errors_summary.csv")

    # Errors.pdf
    errors_figs = ioerrorcalc.make_errors_figs(errors, formatting=rp.PLOT_IV)
    ioerrorcalc.write_errors_pdf(errors_figs, filename = "Errors.pdf")



"""Module cli of viperleed.calc.

Contains the definition of the main interface used for running
viperleed.calc from the command-line as in
>>> python3 -m viperleed.calc
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

import multiprocessing
import os
from pathlib import Path
import shutil

from viperleed.calc import DEFAULT_HISTORY
from viperleed.calc import DEFAULT_WORK
from viperleed.calc import DEFAULT_WORK_HISTORY
from viperleed.calc.bookkeeper import BookkeeperMode
from viperleed.calc.bookkeeper import bookkeeper
from viperleed.calc.lib.base import copytree_exists_ok
from viperleed.calc.lib.leedbase import getMaxTensorIndex
from viperleed.calc.run import run_calc
from viperleed.calc.sections.calc_section import ALL_INPUT_FILES
from viperleed.cli_base import ViPErLEEDCLI


LOG_VERBOSE = 5
LOG_VERY_VERBOSE = 1


class ViPErLEEDCalcCLI(ViPErLEEDCLI, cli_name='calc'):
    """Main command-line interface for viperleed.calc."""

    def __call__(self, args=None):
        """Call viperleed.calc."""
        # Make sure that pre-packed 'executables'
        # work with multiprocessing on Windows
        multiprocessing.freeze_support()

        args = self.parse_cli_args(args)
        work_path = _make_work_directory(args)

        presets = {}  # Replace selected PARAMETERS
        _verbosity_to_log_level(args, presets)

        print('Running bookkeeper...')                                          # TODO: This is lost to stdout if we don't log it
        # NB: job_name is None, as we're cleaning up the previous run
        bookkeeper(mode=BookkeeperMode.DEFAULT,
                   job_name=None,
                   history_name=args.history_name,
                   work_history_name=args.work_history_name)

        _copy_tensors_and_deltas_to_work(work_path, args.all_tensors)           # TODO: it would be nice if all_tensors automatically checked PARAMETERS
        _copy_input_files_to_work(work_path)

        # Go to work directory, execute there
        cwd = Path.cwd().resolve()
        os.chdir(work_path)
        exit_code = 2
        try:
            exit_code = run_calc(
                system_name=args.name,
                source=args.tensorleed,
                preset_params=presets
                )
        finally:
            # Copy back everything listed in manifest, then go back
            _copy_files_from_manifest(cwd)
            os.chdir(cwd)

        # Call bookkeeper again to clean up unless --no-cont is set
        if not args.no_cont:
            bookkeeper(mode=BookkeeperMode.CONT,
                       job_name=args.job_name,
                       history_name=args.history_name,
                       work_history_name=args.work_history_name)

        # Finally clean up work if requested
        if args.delete_workdir:
            try:
                shutil.rmtree(work_path)
            except OSError as exc:
                print(f'Error deleting work directory: {exc}')                  # TODO: This is lost to stdout if we don't log it
        return exit_code

    def add_parser_arguments(self, parser):
        """Add CLI arguments for viperleed.calc to parser."""
        super().add_parser_arguments(parser)
        parser.add_argument(
            '--name', '-n',
            help=('specify name of the system/sample for which '
                  'calculations are run (e.g., Ag(100)-p1x1)'),
            type=str
            )

        # VERBOSITY
        verbosity = parser.add_mutually_exclusive_group()
        verbosity.add_argument(
            '-v', '--verbose',
            help='increase output verbosity and print debug messages',
            action='store_true'
            )
        verbosity.add_argument(
            '-vv', '--very-verbose',
            help='increase output verbosity and print more debug messages',
            action='store_true'
            )

        # PATHS
        parser.add_argument(                                                    # TODO: bookkeeper always assumes DEFAULT_WORK!
            '-w', '--work',
            help='specify execution work directory',
            type=str
            )
        parser.add_argument(
            '--tensorleed', '-t',
            help=('specify the path to the folder containing '
                  'the TensErLEED and EEASISSS source codes'),
            type=str
            )

        # BOOKKEPER
        parser.add_argument(
            '--no-cont',
            help='do not overwrite POSCAR/VIBROCC with those after a search',
            action='store_true'
            )
        parser.add_argument(                                                    # TODO: implement (for cont at end; warn if called with --no_cont)
            '-j', '--job-name',
            help=('define a name for the current run. Will be appended to the '
                  'name of the history folder that is created, and is logged '
                  'in history.info. Passed along to the bookkeeper'),
            type=str
            )
        parser.add_argument(
            '--history-name',
            help=('define the name of the history folder that is '
                  'created/used. Passed along to the bookkeeper. '
                  f'Default is {DEFAULT_HISTORY!r}'),
            type=str,
            default=DEFAULT_HISTORY
            )
        parser.add_argument(
            '--work-history-name',
           help=('define the name of the workhistory folder that '
                 'is created/used. Passed along to the bookkeeper. '
                 f'Default is {DEFAULT_WORK_HISTORY!r}'),
            type=str,
            default=DEFAULT_WORK_HISTORY
            )

        # CREATING/DELETING DIRECTORIES
        parser.add_argument(
            '--all-tensors',
            help=('copy all Tensors to the work directory. Required if using '
                  'the TENSORS parameter to calculate from old tensors'),
            action='store_true'
            )
        parser.add_argument(
            '--delete-workdir',
            help='delete work directory after execution',
            action='store_true'
            )


def _copy_input_files_to_work(work_path):
    """Copy all the known input files present here into work_path."""
    for file in ALL_INPUT_FILES:
        try:
            shutil.copy2(file, work_path / file)
        except FileNotFoundError:
            pass


def _copy_files_from_manifest(to_path):
    """Copy all files listed in file 'manifest' back to_path."""
    manifest_file = Path('manifest')
    if not manifest_file.is_file():
        return

    with manifest_file.open('r', encoding='utf-8') as file:
        manifest = [line.strip() for line in file.readlines()]
        manifest = (line for line in manifest if line)
        manifest = (Path(line) for line in manifest)

    for to_be_copied in manifest:
        _copy = shutil.copy2 if to_be_copied.is_file() else copytree_exists_ok
        try:
            _copy(to_be_copied, to_path / to_be_copied)
        except OSError as exc:
            print(f'Error copying {to_be_copied} to home directory: {exc}')     # TODO: Why no logging?


def _copy_tensors_and_deltas_to_work(work_path, all_tensors):
    """Move appropriate files from 'Tensors' and 'Deltas' to work_path."""
    if all_tensors:  # Copy all of them
        for directory in ('Tensors', 'Deltas'):
            try:
                copytree_exists_ok(directory, work_path / directory)
            except FileNotFoundError:
                pass
        return

    # Only copy the most recent Tensors/Deltas from zip files
    tensor_num = getMaxTensorIndex(zip_only=True)
    if not tensor_num:
        return

    for local_dir in ('Tensors', 'Deltas'):
        local_file = Path(f'{local_dir}/{local_dir}_{tensor_num:03d}.zip')
        if not local_file.is_file():
            continue
        work_directory = work_path / local_dir
        work_directory.mkdir(parents=True, exist_ok=True)
        shutil.copy2(local_file, work_path / local_file)


def _make_work_directory(cli_args):
    """Return a suitable 'work' directory from cli_args."""
    work_path = Path(cli_args.work or DEFAULT_WORK).resolve()
    work_path.mkdir(parents=True, exist_ok=True)
    return work_path


def _verbosity_to_log_level(cli_args, presets):
    """Add a LOG_LEVEL to presets if cli_args have verbosity specified."""
    if cli_args.very_verbose:
        presets['LOG_LEVEL'] = LOG_VERY_VERBOSE
    elif cli_args.verbose:
        presets['LOG_LEVEL'] = LOG_VERBOSE



"""Package lib of viperleed.calc.

Defines generic functionality used throughout the viperleed.calc
package.

Modules
-------
base
    Non-LEED-related classes and functions.
checksums
    Functions used to prevent user tinkering with FORTRAN source files.
    The actual checksums are contained in _checksums.dat and should
    only be edited by expert contributors.
leedbase
    LEED- and TensErLEED-related base functions used in calc.
periodic_table
    Collection of basic atom data.
"""
# NB: woods_notation is undocumented on purpose, as it will be
# replaced with functionality in guilib.

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-21'
__license__ = 'GPLv3+'



"""Package rparams of viperleed.calc.classes.

This package is a result of the refactor of the rparams.py
module, originally by @fkraushofer. Contains definition of
Rparams, the class containing most of the information during
a viperleed.calc run. It also defines specific classes for
not-so-simple user parameters (in package special), used
as attributes of Rparams. Defaults for the parameters are
defined in module defaults. Their variability limits in
module limits. Some of the special parameters define their
own defaults and limits.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-23'
__license__ = 'GPLv3+'

from .domain_params import DomainParameters
from .rparams import Rparams
from .special.energy_range import EnergyRange, IVShiftRange, TheoEnergies
from .special.l_max import LMax
from .special.layer_cuts import LayerCuts
from .special.search_cull import SearchCull
from .special.symmetry_eps import SymmetryEps



"""Functions for reading and writing the PHASESHIFTS file."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-18'
__license__ = 'GPLv3+'

from itertools import combinations
import logging
import os
from pathlib import Path

import fortranformat as ff
import numpy as np

from viperleed.calc.lib.leedbase import HARTREE_TO_EV
from viperleed.calc.lib.periodic_table import (get_atomic_number,
                                               get_element_symbol)

try:
    import matplotlib
except ImportError:
    _CAN_PLOT = False
else:
    matplotlib.rcParams.update({'figure.max_open_warning': 0})
    matplotlib.use('Agg')
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.pyplot as plt
    plt.style.use('viperleed.calc')
    _CAN_PLOT = True


logger = logging.getLogger(__name__)


def readPHASESHIFTS(sl, rp, readfile='PHASESHIFTS', check=True,
                    ignoreEnRange=False):
    """Read from a PHASESHIFTS file.

    Parameters
    ----------
    sl: Slab
    rp: RunParameters
    readfile: str, optional
        filename to be read. Default is 'PHASESHIFTS'.
    check: bool, optional
        Wether to check for consistence agains sl and rp.
        If False, sl and rp can be None. Default is True.
    ignoreEnRange: bool, optional
        Check wether the energy range in readfile is sufficient to
        cover the energy range requested in rp. Default is False.

    Returns
    -------
    firstline: str
        Header line of the readfile, containing Rundgren parameters.
    phaseshifts: list of tuple
        Each element is (energy, phaseshifts_at_energy) with
        phaseshifts_at_energy = [[el0_L0, el0_L1, ...], [el1_L0, ...], ...]
        where eli_Lj is the phaseshift for element i and angular momentum j.
        Therefore, len(phaseshifts) is the number of energies found,
        len(phaseshifts[0][1]) should match the number of sites, and
        len(phaseshifts[0][1][0]) is the number of different
        values of L in the phaseshift file.
    newpsGen: bool
        Whether the inconsitency found requires a full recalculation
        of the phaseshifts. This happens if:
            1) an error occurs while parsing,
            2) if check, rp.V0_real was not given and could not
               interpret firstline,
            3) LMAX of readfile is smaller than required in rp,
            4) if number of blocks inconsitent with elements in sl.
    newpsWrite: bool
        Whether the inconsitency found requires writing a new
        PHASESHIFTS file. This is distinct from newpsGen as we
        can at times generate a new file from the information read.
        This is the case if:
            5) There are fewer blocks than expected, but the number of blocks
                matches the number of chemical elements in sl. This means however,
                that all sites with the same chemical element will use the same
                phaseshift.
    """
    rf74x10 = ff.FortranRecordReader('10F7.4')
    ri3 = ff.FortranRecordReader('I3')

    # legacy - allow "_PHASESHIFTS"
    if (readfile == 'PHASESHIFTS' and not os.path.isfile('PHASESHIFTS')
            and os.path.isfile('_PHASESHIFTS')):
        logger.info("Found no PHASESHIFTS file, but found legacy file named "
                    "_PHASESHIFTS. Renaming _PHASESHIFTS to PHASESHIFTS.")
        os.rename('_PHASESHIFTS', 'PHASESHIFTS')

    try:
        rf = open(readfile, 'r')
    except FileNotFoundError:
        logger.error("PHASESHIFTS file not found.")
        raise
    else:
        filelines = rf.readlines()
    finally:
        rf.close()

    try:
        nel = ri3.read(filelines[0])[0]
    except (ValueError, IndexError):
        logger.error("Exception while trying to read PHASESHIFTS: could not "
                     "find number of blocks in first line.")
        raise
    phaseshifts = []

    firstline = filelines[0]
    line_idx = 1
    try:
        lines_per_block = find_block_size(filelines[1:])
    except ValueError as err:
        logger.warning(
            f"Error while trying to read PHASESHIFTS: {err} "
            "A new PHASESHIFTS file will be generated."
        )
        rp.setHaltingLevel(1)
        return "", [], True, True

    # check block length is consitent with number of species
    if (lines_per_block - 1) % nel:
        logger.warning(
            "Error while trying to read PHASESHIFTS: "
            "Could not parse file: The number of blocks may not match "
            "the number given in the first line. A new PHASESHIFTS "
            "file will be generated."
        )
        rp.setHaltingLevel(1)
        return "", [], True, True

    lines_per_element = (lines_per_block - 1) // nel

    while line_idx < len(filelines):
        current_line = filelines[line_idx].strip()
        if not current_line:
            line_idx += 1
            continue
        energy = rf74x10.read(filelines[line_idx])[0] # read energy
        enps = []
        for i in range(nel):
            elps = []
            for j in range(lines_per_element):
                llist = rf74x10.read(filelines[line_idx
                                                + (i*lines_per_element)+j+1])
                llist = [f for f in llist if f is not None]
                elps.extend(llist)
            enps.append(elps)
        phaseshifts.append((energy, enps))
        line_idx += lines_per_element*nel+1

    if not check:
        return firstline, phaseshifts, False, False

    # try extracting muffin tin parameters:
    muftin = []
    llist = firstline.split()
    if len(llist) >= 6:
        for i in range(1, 5):
            try:
                muftin.append(float(llist[i]))
            except (ValueError, IndexError):
                muftin = []
                break
    else:
        muftin = []

    (phaseshifts,
     firstline,
     newpsGen,
     newpsWrite) = __check_consistency_rp_elements(sl, rp, phaseshifts,
                                                   firstline, muftin)

    if not ignoreEnRange and not newpsGen:
        newpsGen = __check_consistency_energy_range(rp, phaseshifts,
                                                    muftin, newpsGen)
        newpsWrite |= newpsGen

    # Check consitency of phaseshift values (unless we
    # already know we have to make a new file anyway)
    if not newpsGen:
        __check_consitency_element_order(rp, sl, phaseshifts)

    return firstline, phaseshifts, newpsGen, newpsWrite


def __check_consistency_rp_elements(sl, rp, phaseshifts, firstline, muftin):
    """Check that phaseshifts fit the expected number of elements and LMAX.

    Parameters
    ----------
    sl : Slab
        Surface Slab object.
    rp : Rparams
        Run paramters.
    phaseshifts : list
        Nested list of phaseshifts.
    firstline : str
        First line of the phaseshifts file.
    muftin : list
        Rundgren coefficients interpreted from the firstline.

    Returns
    -------
    phaseshifts: list of tuple
        Each element is (energy, phaseshifts_at_energy) with
        phaseshifts_at_energy = [[el0_L0, el0_L1, ...], [el1_L0, ...], ...]
        where eli_Lj is the phaseshift for element i and angular momentum j.
        Therefore, len(phaseshifts) is the number of energies found,
        len(phaseshifts[0][1]) should match the number of sites, and
        len(phaseshifts[0][1][0]) is the number of different
        values of L in the phaseshift file.
    firstline: str
        Header line of the file read, containing Rundgren parameters,
        modified if newpsWrite is True.
    newpsGen: bool
        Wether the inconsitency found requires a full recalculation
        of the phaseshifts. This happens if:
            1) rp.V0_real was not given and could not interpret firstline,
            2) LMAX in phaseshifts is smaller than required in rp,
            3) if number of blocks inconsitent with elements in sl.
    newpsWrite: bool
        Wether the inconsitency found requires writing a new
        PHASESHIFTS file. This is distinct from newpsGen as we
        can at times generate a new file from the information read.
        This is the case if:
            4) There are fewer blocks than expected, but the number of
               blocks matches the number of chemical elements in sl.
               This means however, that all sites with the same element
               will use the same phaseshift.
    """
    newpsGen, newpsWrite = True, True

    n_l_values = len(phaseshifts[0][1][0])
    n_el_and_sites = len(phaseshifts[0][1])

    n_el_and_sites_expected = 0
    for el in sl.elements:
        if el in rp.ELEMENT_MIX:
            n = len(rp.ELEMENT_MIX[el])
        else:
            n = 1
        n_el_and_sites_expected += n*len([s for s in sl.sitelist if s.el == el])

    if rp.V0_REAL == "default" and not muftin:
        logger.warning(
            "Could not convert first line of PHASESHIFTS file to MUFTIN "
            "parameters. A new PHASESHIFTS file will be generated."
            )
        rp.setHaltingLevel(1)

    elif n_el_and_sites == n_el_and_sites_expected:
        logger.debug(f"Found {n_el_and_sites_expected} blocks in PHASESHIFTS "
                     "file, which is consistent with PARAMETERS.")
        newpsGen, newpsWrite = False, False

    # Check that the phaseshifts read in have sufficient lmax
    elif n_l_values < rp.LMAX.max + 1:  # +1 because of L=0
        logger.warning(
            "Maximum angular momentum LMAX in PHASESHIFTS "
            "file is lower than required by PARAMETERS. A "
            "new PHASESHIFTS file will be generated."
            )
        rp.setHaltineLevel(1)

    elif n_el_and_sites == len(sl.chemelem):
        logger.warning(
            "Found fewer blocks than expected in the "
            "PHASESHIFTS file. However, the number of blocks matches "
            "the number of chemical elements. A new PHASESHIFTS file "
            "will be generated, assuming that each block in the old "
            "file should be used for all atoms of one element."
            )
        rp.setHaltingLevel(1)
        oldps = phaseshifts[:]
        phaseshifts = []
        for (energy, oldenps) in oldps:
            enps = []
            j = 0   # block index in old enps
            for el in sl.elements:
                if el in rp.ELEMENT_MIX:
                    m = len(rp.ELEMENT_MIX[el])
                else:
                    m = 1
                n = len([s for s in sl.sitelist if s.el == el])
                for i in range(0, m):    # repeat for chemical elements
                    for k in range(0, n):   # repeat for sites
                        enps.append(oldenps[j])
                    j += 1  # count up the block in old enps
            phaseshifts.append((energy, enps))
        newpsGen = False
        firstline = str(len(phaseshifts[0][1])).rjust(3) + firstline[3:]

    else:
        logger.warning(
            "PHASESHIFTS file was read but is inconsistent with "
            "PARAMETERS. A new PHASESHIFTS file will be generated."
            )
        rp.setHaltingLevel(1)

    return phaseshifts, firstline, newpsGen, newpsWrite


def __check_consistency_energy_range(rp, phaseshifts, muftin, newpsGen):
    """Check that the energy range of phaseshifts is large enough for rp."""
    checkfail = False

    er = np.arange(rp.THEO_ENERGIES.start,
                   rp.THEO_ENERGIES.stop + 1e-4,
                   rp.THEO_ENERGIES.step)
    psmin = round(phaseshifts[0][0] * HARTREE_TO_EV, 2)
    psmax = round(phaseshifts[-1][0] * HARTREE_TO_EV, 2)

    if rp.V0_REAL == "default" or isinstance(rp.V0_REAL, list):
        if isinstance(rp.V0_REAL, list):
            c = rp.V0_REAL
        else:
            c = muftin
            checkfail = not bool(muftin)
        if c and not checkfail:
            # energies at which scattering occurs
            er_inner = er + rp.FILAMENT_WF
            er_inner -= np.maximum(
                c[0],
                c[1] + (c[2]/np.sqrt(er_inner + c[3]))
                )
    else:
        try:
            er_inner = er + float(rp.V0_REAL)
        except (ValueError, TypeError):
            checkfail = True

    if checkfail:
        logger.warning(
            "Could not check energy range in PHASESHIFTS "
            "file. If energy range is insufficient, try deleting the "
            "PHASESHIFTS file to generate a new one."
            )
        return newpsGen

    if (psmin > min(er_inner) or psmax < max(er_inner)):
        if (psmin > min(er_inner) and psmin <= 20.
                and psmax >= max(er_inner)):
            # can lead to re-calculation of phaseshifts every run if
            # V0r as calculated by EEASiSSS differs from 'real' V0r.
            # Don't automatically correct.
            logger.warning(
                f"Lowest value in PHASESHIFTS file ({psmin:.1f} eV) is "
                "larger than the lowest predicted scattering energy "
                f"({min(er_inner):.1f} eV). If this causes problems in the "
                "reference calculation, try deleting the PHASESHIFTS "
                "file to generate a new one, or increase the starting "
                "energy in the THEO_ENERGIES parameter."
                )
        else:
            logger.warning(
                "The energy range found in the PHASESHIFTS "
                "file is smaller than the energy range requested for "
                "theoretical beams. A new PHASESHIFTS file will be "
                "generated."
                )
            return True

    return newpsGen


def __check_consitency_element_order(rp, sl, phaseshifts,
                                     eps=None, l_max_cutoff=4):
    """Determine if elements may have been assigned wrong phaseshifts.

    In general at high energies and at high LMAX, heavier elements
    (higher atomic number) have larger elastic scattering phaseshifts
    than lighter atoms. If this is not the case in the read in
    PHASESHIFTS file, the ordering in the file may be wrong.
    This is a very common user mistake, that often occurs when
    using a different POSCAR with pre-existing PHASESHIFTS.

    Warnings will be issued if an inconsistency is found for
    elements whose atomic numbers differ by at least 2. This
    means, e.g., that Cr--Ni will not issue warnings for Fe,
    but Ti and Cu will.

    Parameters
    ----------
    rp : Rparams
        Run parameters.
    sl : Slab
        Surface Slab object. Contains site types to check against.
    phaseshifts : list
        Phaseshifts list, as read in from a PHASESHIFTS file.
    eps : float, optional
        Tolerance to use when comparing phaseshift values. If None
        or not given, eps == rp.PHASESHIFT_EPS. Default is None.
    l_max_cutoff : int, optional
        Lower cutoff for angular momentum quantum number to be used for
        comparison of phaseshifts. Heavier elements should always have
        higher phaseshifts in the limit of high energy and high angular
        momentum. Behaviour for low energy/low angular momentum is not
        as clear cut, as phaseshifts may cross zero and are pi periodic.
        Default is 4.

    Returns
    -------
    may_have_wrong_phaseshifts : set
        Set of chemical elements (str) which may be assigned the wrong
        phaseshifts. If no inconsitency is detected, the set is empty.
    """
    if not eps:
        eps = rp.PHASESHIFT_EPS
    n_l_values = len(phaseshifts[0][1][0])
    n_sites_in_ps = len(phaseshifts[0][1])  # No. species in PHASESHIFTS

    # Get atomic number of all site types in the order
    # they appear in sl.sitelist, taking into account
    # that some sites may have mixed occupation
    element_mix = []
    for site in sl.sitelist:
        element_mix.extend(
            site.mixedEls  # Mixed occupation
            or [site.el]   # Only one element
            )
    # get atomic numbers from element symbols
    real_el = lambda el: (rp.ELEMENT_RENAME[el] if el in rp.ELEMENT_RENAME
                          else el)
    atomic_numbers = tuple(get_atomic_number(real_el(el))
                           for el in element_mix)

    # Get phaseshifts at highest LMAX that are larger than eps
    en_idx = -1  # look at highest energy only
    for l_value in reversed(range(l_max_cutoff, n_l_values)):
        ps_sites = np.array(list(
            phaseshifts[en_idx][1][site_idx][l_value]
            for site_idx in range(n_sites_in_ps)
            ))
        if all(abs(ps_sites) > eps):
            break
    else:
        # None of the phaseshifts are larger than eps at this energy
        logger.warning(
            "Could not check consistency of PHASESHIFTS file: All "
            f"PHASESHIFTS are smaller than {eps} at the largest energy."
            )
        rp.setHaltingLevel(1)
        return

    affected_elements = set()
    ps_pairs = list(zip(atomic_numbers, ps_sites))
    at_number = lambda item : item[0]
    pair_ps = lambda item : item[1]
    ps_pairs.sort(key=at_number) # sort by atomic number

    # Go through all pairs and check that heavier
    # elements have larger phaseshifts
    for pair_1, pair_2 in combinations(ps_pairs, 2):
        at_num_change = at_number(pair_1) - at_number(pair_2)
        if at_num_change <= 0:  # Same element or redundant iteration
            continue
        # Different element. Check that phaseshifts are also
        # ordered, and complain only if elements are not
        # close neighbors
        if (abs(pair_ps(pair_1)) >= abs(pair_ps(pair_2)) - eps
                and at_num_change > 2):
            continue
        affected_elements.update((at_number(pair_1), at_number(pair_2)))
    # get element symbols
    may_have_wrong_phaseshifts = set(get_element_symbol(el)
                                     for el in affected_elements)

    if may_have_wrong_phaseshifts:
        elements_str = ", ".join(affected_elements)
        logger.warning(
            "Detected inconsitency in PHASESHIFTS: "
            "Found larger phaseshifts for lighter elements "
            f"than for heavier elements at LMAX={l_value}. "
            "This may mean that the blocks in PHASESHIFTS "
            "are in the wrong order! "
            f"Affected elements: {elements_str}. "
            "You may want to regenerate the PHASESHIFTS file or "
            "reorder the blocks (see utility ModifyPhaseshifts). "
        )
        rp.setHaltingLevel(1)
    return may_have_wrong_phaseshifts


def writePHASESHIFTS(firstline, phaseshifts, file_path=Path()/'PHASESHIFTS'):
    """Takes phaseshift data and writes it to a PHASESHIFTS file."""
    _file_path = Path(file_path)
    output = firstline
    if output[-1] != "\n":
        output += "\n"
    f74x10 = ff.FortranRecordWriter('10F7.4')
    f74 = ff.FortranRecordWriter('F7.4')
    for (en, enps) in phaseshifts:
        output += f74.write([en])+"\n"
        for block in enps:
            output += f74x10.write(block)+"\n"
    try:
        with open(_file_path, 'w') as wf:
            wf.write(output)
        logger.debug(f"Wrote to {_file_path} successfully.")
    except Exception:
        logger.error("Exception while writing PHASESHIFTS file: ",
                     exc_info=True)
        raise
    return


def plot_phaseshifts(sl, rp, filename="Phaseshifts_plots.pdf"):
    """
    Outputs plots of the phaseshifts in rp.phaseshifts in a pdf file.

    Parameters
    ----------
    sl : Slab
        Slab object, used to determine elements and sites represented in
        the phaseshifts.
    rp : Rparams
        Run parameters. Phaseshifts are read from rp.phaseshifts.
    filename : str, optional
        Name of the output file. The default is "Phaseshifts_plots.pdf".

    Returns
    -------
    None.
    """
    if not _CAN_PLOT:
        logger.debug("Necessary modules for plotting not found. Skipping "
                     "Phaseshift plotting.")
        return
    ps_labels = []
    for el in sl.elements:
        # this reproduces the order of blocks contained in PHASESHIFTS:
        if el in rp.ELEMENT_MIX:
            chemelList = rp.ELEMENT_MIX[el]
        elif el in rp.ELEMENT_RENAME:
            chemelList = [rp.ELEMENT_RENAME[el]]
        else:
            chemelList = [el]
        ps_labels.extend([cel + " in " + s.label + " site"
                          for cel in chemelList
                          for s in sl.sitelist if s.el == el])
    energies = np.array([ps[0]*HARTREE_TO_EV for ps in rp.phaseshifts])
    ps_vals = np.array([ps[1] for ps in rp.phaseshifts])

    figsize = (7, 4)
    linewidth = 1
    nlplot = min(np.shape(ps_vals)[-1], rp.LMAX.max+1)

    figs = []
    # colors = ["#000000", "#004949", "#009292", "#ff6db6", "#ffb6db",
    #           "#490092", "#006ddb", "#b66dff", "#6db6ff", "#b6dbff",
    #           "#920000", "#924900", "#db6d00", "#24ff24", "#ffff6d"]
    #   colorblind safe 16 - not great
    colors = ["#000000", "#E69F00", "#56B4E9", "#009E73",
              "#F0E442", "#0072B2", "#D55E00", "#CC79A7"]  # colorblind safe 8
    styles = ["-", "--", ":", "-."]  # for when we run out of colors
    for i, label in enumerate(ps_labels):
        fig, ax = plt.subplots(figsize=figsize)
        figs.append(fig)
        fig.subplots_adjust(left=0.1, right=0.98,
                            bottom=0.11, top=0.92)
        ax.set_title(label)
        ax.set_xlabel("Energy (eV)")
        ax.set_ylabel("Phaseshifts (rad)")
        ax.grid(True, linewidth=0.1*linewidth)
        [sp.set_linewidth(0.7*linewidth) for sp in ax.spines.values()]
        ax.tick_params(bottom=True, top=True, left=True, right=True,
                       axis='both', direction='in', width=0.7*linewidth)
        ax.set_xlim((np.min(energies), np.max(energies)))
        for j in range(0, nlplot):
            ax.plot(energies, ps_vals[:, i, j],
                    linewidth=linewidth, label=fr"$\ell$ = {j}",
                    c=colors[j % len(colors)],
                    ls=styles[(j // len(colors)) % len(styles)])
        legend = ax.legend(ncol=(nlplot // 8 + 1))
        legend.get_frame().set_linewidth(linewidth*0.7)

    try:
        pdf = PdfPages(filename)
        for fig in figs:
            pdf.savefig(fig)
            plt.close(fig)
        pdf.close()
    except PermissionError:
        logger.error("plot_phaseshifts: Cannot open file {}. Aborting."
                     .format(filename))
    return


def find_block_size(filelines):
    """Returns the periodicity of lengths found in filelines."""
    lens_as_chars = "".join(chr(len(l.rstrip())) for l in filelines)

    # Allow empty lines at the end
    while lens_as_chars.endswith(chr(0)):
        lens_as_chars = lens_as_chars[:-1]
    if chr(0) in lens_as_chars:
        raise ValueError("Empty line found.")

    # See stackoverflow.com/questions/29481088
    period = (lens_as_chars * 2).find(lens_as_chars, 1, -1)
    if period == -1:
        raise ValueError("Could not identify block.")

    return period



"""ViPErLEED utility: Sort slab by z."""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from viperleed.utilities.poscar.base import _PoscarStreamCLI


class SortByZCLI(_PoscarStreamCLI, cli_name='sort_by_z'):
    """Sort atoms by ascending or descending z coordinate."""

    long_name = 'sort slab by z'

    def add_parser_arguments(self, parser):
        """Add optional --reversed argument."""
        super().add_parser_arguments(parser)
        parser.add_argument(
            '-r', '--reversed',
            help='sort from bottom to top. Default is from top to bottom.',
            action='store_true',
            )

    def process_slab(self, slab, args):
        """Return a slab z-sorted according to args."""
        slab.sort_by_z(bottom_to_top=args.reversed)
        return slab


if __name__ == '__main__':
    SortByZCLI.run_as_script()



"""Module cli of ViPErLEED POSCAR utilities.

Defines the main-entry-point function for the POSCAR utilities
and other command-line related functionality.

The functionality in this module used to be in __main__.py.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-04'
__license__ = 'GPLv3+'


from viperleed.cli_base import ViPErLEEDCLIWithAutoChildren


class PoscarUtilsCLI(ViPErLEEDCLIWithAutoChildren,
                     cli_name='poscar',
                     help_='utilities for POSCAR files'):
    """Command-line interface for all POSCAR utilities."""

    def add_parser_arguments(self, parser):
        """Add POSCAR arguments to `parser`."""
        self.add_verbose_option(parser)
        super().add_parser_arguments(parser)



"""
======================================
  ViPErLEED Graphical User Interface
======================================
    *** module guilib.leedsim ***

Schematic simulation of a LEED pattern. Allow exporting list of LEED spots for
properly indexing an experimental LEED pattern
"""

from viperleed import GLOBALS
# from vprglobals import GLOBALS

__all__ = ['LEEDSymmetryDomains', 'LEEDsubpattern', 'Woods', 'LEEDPattern',
           'RealSpace', 'export_pattern_csv', 'LEEDParameters',
           'LEEDParametersList', 'LEEDEquivalentBeams', 'LEEDStructuralDomains']

from viperleed.guilib.leedsim.classes import (LEEDEquivalentBeams,    # not sure I need this to be exposed
                                              LEEDSymmetryDomains,    # not sure I need this to be exposed
                                              LEEDsubpattern, Woods,  # not sure I need these to be exposed, there were also mpl_colors and degrees
                                              LEEDPattern, RealSpace,
                                              LEEDStructuralDomains)  # not sure I need this to be exposed
from viperleed.guilib.leedsim.exportcsv import export_pattern_csv
from viperleed.guilib.leedsim.leedparameters import (LEEDParameters,
                                                     LEEDParametersList)
# from .classes import *
# from .exportcsv import export_pattern_csv
# from .leedparameters import *
if GLOBALS['USE_GUI']:
    from viperleed.guilib.leedsim.widgets import (DomsBlock, EnergyBlock,
                                                  RotationBlock, ToggleButton,
                                                  LEEDCanvas, RealCanvas,
                                                  HoverAnnot, MatricesPopup,
                                                  TEST)                            # This probably not...
    from viperleed.guilib.leedsim.mainwindow import LEED_GUI
    from viperleed.guilib.leedsim.mainwindow import show_use_betatest_version_popup
    from viperleed.guilib.leedsim.dialogbulk3dsym import Bulk3DSymDialog
    from viperleed.guilib.leedsim.newfiledialog import NewFileDialog
    from viperleed.guilib.leedsim.exportcsvdialog import ExportCSVDialog
    # from .widgets import *
    # from .mainwindow import LEED_GUI
    # from .NewFileDialog import NewFileDialog
    # from .ExportCSVDialog import ExportCSVDialog
    __all__.extend(['DomsBlock', 'EnergyBlock', 'RotationBlock', 'ToggleButton',
                    'show_use_betatest_version_popup',
                    'LEEDCanvas', 'RealCanvas', 'HoverAnnot', 'MatricesPopup',
                    'TEST', 'LEED_GUI', 'Bulk3DSymDialog', 'NewFileDialog',
                    'ExportCSVDialog'])





"""ViPErLEED utility: Find Symmetry."""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from viperleed.calc import symmetry
from viperleed.utilities.poscar.base import _PoscarSymmetryCLI


class FindSymmetryCLI(_PoscarSymmetryCLI, cli_name='find_symmetry'):
    """Detect the plane group of a the structure in a POSCAR."""

    long_name = 'find symmetry'

    def process_slab(self, slab, args):
        """Find the plane group of slab."""
        rpars = self.prepare_rpars(slab, args)
        symmetry.findSymmetry(slab, rpars)

    def write_output(self, processed_slab, args):
        """Write the detected plane group to the args.outfile."""
        args.outfile.write(f'{processed_slab.foundplanegroup}\n')


if __name__ == '__main__':
    FindSymmetryCLI.run_as_script()



"""Module l_max of viperleed.calc.classes.rparams.special.

Defines the LMax class, a convenience container for parameter LMAX.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-12-16'
__license__ = 'GPLv3+'

from dataclasses import dataclass

from ..defaults import NO_VALUE
from ..limits import PARAM_LIMITS
from .base import SpecialParameter


_MIN, _MAX = PARAM_LIMITS['LMAX']


@dataclass(repr=False)
class LMax(SpecialParameter, param='LMAX'):
    """A container for user-supplied LMAX values."""

    _min : int = NO_VALUE
    _max : int = NO_VALUE

    def __post_init__(self):
        """Process the input values to integers."""
        for attr_name in ('_min', '_max'):
            attr = getattr(self, attr_name)
            value = self._check_int_in_range(attr)
            setattr(self, attr_name, value)
        self._sort_values()

    def __bool__(self):
        """Return whether any of the LMax values is defined."""
        return self.has_max or self.has_min

    def __repr__(self):
        """Return a representation string of this LMax."""
        txt = 'LMax('
        if not self:
            return txt + ')'
        txt += str(self.min)  # == .max if only one value
        if not self.has_single_value:
            txt += f', {self.max}'
        return txt + ')'

    def __str__(self):
        """Return a string version of this LMax."""
        if not self:
            return ''
        if self.has_single_value:
            return str(self.max)
        return f'{self.min}-{self.max}'

    @staticmethod
    def _check_int_in_range(value):
        """Return an in-range integer version of value. Raise otherwise."""
        if value is NO_VALUE:
            return value
        try:
            f_value = float(value)
        except (TypeError, ValueError):
            raise TypeError('LMax must be numeric') from None
        i_value = int(f_value)
        if abs(i_value - f_value) > 1e-6:
            raise TypeError('LMax must be integer')
        if not _MIN <= i_value <= _MAX:
            raise ValueError(f'Invalid value={value} for LMax. Values '
                             f'must be between {_MIN} and {_MAX}')
        return i_value

    def _sort_values(self):
        """Swap min and max so that they're sorted as expected."""
        if not self or self.has_single_value:
            return
        if self._min > self._max:
            self._min, self._max = self._max, self._min

    @property
    def has_min(self):
        """Return whether there is a defined minimum value."""
        return self._min is not NO_VALUE

    @property
    def has_max(self):
        """Return whether there is a defined maximum value."""
        return self._max is not NO_VALUE

    @property
    def has_single_value(self):
        """Return whether this LMax does not correspond to a range."""
        return bool(self) and self.min == self.max

    @property
    def max(self):
        """Return the largest value of LMAX to be used for calculations."""
        return self._max if self.has_max else self._min

    @max.setter
    def max(self, new_max):
        """Assign a new maximum value for this LMax."""
        value = self._check_int_in_range(new_max)
        self._max = value
        self._sort_values()

    @property
    def min(self):
        """Return the smallest value of LMAX to be used for calculations."""
        return self._min if self.has_min else self._max

    @min.setter
    def min(self, new_min):
        """Assign a new minimum value of this LMax."""
        value = self._check_int_in_range(new_min)
        self._min = value
        self._sort_values()

    @classmethod
    def from_value(cls, value):
        """Return an LMax from a 2-tuple value."""
        return cls(*value)



"""Package parameters of viperleed.calc.files.

Contains definitions useful for reading, writing, editing, and
interpreting a PARAMETERS file. This package comes from a refactor
of the former parameters.py module.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-13'
__license__ = 'GPLv3+'

from . import errors
from .interpret import ParameterInterpreter
from .interpret import interpret
from .read import read
from .read import update
from .write import comment_out
from .write import modify



"""Module cli_base of viperleed.

Defines classes and functions that can (and should) be used to add new
command-line utilities to viperleed.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-26'
__license__ = 'GPLv3+'

import argparse
from collections import defaultdict
from functools import partial
import importlib
import inspect
import logging
from pathlib import Path
import pkgutil
import sys

from viperleed import GLOBALS
from viperleed.calc.lib.base import parent_name


# Default name of the submodules of package utilities
CLI_MODULE_NAME = 'cli'


class NotAViPErLEEDCLIError(Exception):
    """Object is not a ViPErLEEDCLI, or module does not define one."""


def _to_float(value_str):
    """Raise ArgumentTypeError if `value_str` is not a float."""
    try:
        return float(value_str)
    except ValueError:
        raise argparse.ArgumentTypeError('Must be a floating-point '
                                         'value') from None


def float_in_zero_one(value_str):
    """Raise ArgumentTypeError if `value_str` is not in range (0, 1).

    This can be used as a type checker for parser.add_argument.

    Parameters
    ----------
    value_str : str
        The raw value supplied via the command-line.

    Raises
    ------
    ArgumentTypeError
        If `value_str` cannot be converted to float or if it is
        not in range (0, 1).
    """
    value = _to_float(value_str)
    if not 0 < value < 1:
        raise argparse.ArgumentTypeError('Must be between zero '
                                         'and one (excluded)')
    return value


def length_choices(*choices):
    """Return an action that forces `choices` as the only acceptable nargs."""

    if len(choices) > 2:  # pylint: disable=magic-value-comparison
        fmt_choices = ', '.join(str(c) for c in choices[:-1])
        fmt_choices += f', or {choices[-1]}'
    else:
        fmt_choices = ' or '.join(str(c) for c in choices)

    # pylint: disable-next=too-few-public-methods
    class _LengthChoices(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            n_values = len(values)
            if n_values not in choices:
                err_ = f'argument {self.dest!r} requires {fmt_choices} values'
                parser.error(err_)
            setattr(args, self.dest, values)

    return _LengthChoices


def positive_float(value_str):
    """Raise ArgumentTypeError if `value_str` is not strictly positive.

    This can be used as a type checker for parser.add_argument.

    Parameters
    ----------
    value_str : str
        The raw value supplied via the command-line.

    Raises
    ------
    ArgumentTypeError
        If `value_str` cannot be converted to float or if it is <= 0.
    """
    value = _to_float(value_str)
    if value <= 0:
        raise argparse.ArgumentTypeError('Must be strictly positive')
    return value


def required_length(n_min=None, n_max=None):
    """Return an action ensuring there are between n_min and n_max items."""
    # Code adapted from https://stackoverflow.com/questions/4194948

    # pylint: disable-next=too-few-public-methods
    class _RequiredLength(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            n_values = len(values)
            err_ = f'argument {self.dest!r} requires '
            if n_min is not None and n_values < n_min:
                err_ += f'at least {n_min} values'
                parser.error(err_)
            if n_max is not None and n_values > n_max:
                err_ += f'at most {n_max} values'
                parser.error(err_)
            setattr(args, self.dest, values)

    return _RequiredLength


minimum_length = partial(required_length, n_max=None)
maximum_length = partial(required_length, n_min=None)


def strip_cli_module(module_name):
    """Return `module_name` without a terminating '.cli' (if it has one)."""
    return parent_name(module_name, remove=CLI_MODULE_NAME)


class ViPErLEEDCLI:
    """Base class for any viperleed command-line interface.

    How to create a new viperleed CLI/utility
    -----------------------------------------
    Subclass `ViPErLEEDCLI`, optionally passing a name.
        The subclass should be defined in a script-like module (small
        single-module utilities) or in a cli.py module of a utility
        package. Subclasses whose name starts with an underscore are
        considered to be base implementations to be further extended
        and will not be collected as valid children. Normally, each
        utility module should define only a single ViPErLEEDCLI
        subclass. For info, see
        >>> help(ViPErLEEDCLI.__init_subclass__)
    If needed, override the __call__ method.
        This should be the main functionality of the CLI tool. For
        detailed info, see
        >>> help(ViPErLEEDCLI.__call__)
    Extend the `add_parser_arguments` method if the utility has special
        arguments. For detailed info, see
        >>> help(ViPErLEEDCLI.add_parser_arguments)
    Checking of command-line arguments should be done in
        `parse_cli_args`, and, only in exceptional cases
        in `__call__`. For detailed info, see
        >>> help(ViPErLEEDCLI.parse_cli_args)
    A utility can have children sub-utilities. If this is the case,
        call `register_child`, typically in `__init__`. Notice that
        the order in which children are registered is the one with
        which they will later appear. To automatically filter out
        invalid modules/packages (i.e., those that do not define a
        subclass of `ViPErLEEDCLI`) use `register_valid_children.`
        See also the `ViPErLEEDCLIWithAutoChildren` subclass of this
        module that automatically finds all valid children.
    Children can be given aliases too (with `add_child_aliases`).
        They will appear as alternative command names.
    """

    @classmethod
    def __init_subclass__(cls, *,
                          cli_name=None,
                          description=None,
                          help_=None,
                          **kwargs):
        """Create a ViPErLEED command-line interface subclass.

        Parameters
        ----------
        cli_name : str, optional
            Typically the __name__ of the module that implements the
            children utility. If not given, cls.__module__ (stripped
            of the last .cli entry) is used instead. Default is None.
        description : str, optional
            Description text that will be added to the CLI.
            Default is None.
        help_ : str, optional
            Help text to be displayed when this utility is used as a
            child of another one. If not given, a default help text
            is generated from `cli_name`. Default is None.
        **kwargs : dict
            Other optional arguments, passed on to type.

        Returns
        -------
        None.
        """
        super().__init_subclass__(**kwargs)
        # For each class, store a {module_name: cls} register of all
        # the ViPErLEEDCLI classes that would like to be a child
        # of this utility. module_name is the cli_name of the module
        # in which the CLI utility is defined. The cli_name excludes,
        # however a terminating '.cli', as sub-utilities should
        # normally define their command-line interface either in
        # the main module itself (for a single-module utility) or
        # in a cli.py module.
        # Notice that each new subclass MUST have its own dedicated
        # _children dictionary to avoid infinite recursion when
        # sub-classing twice the same subclass of ViPErLEEDCLI.
        cls._children = {}
        cls._cli_name = cli_name or strip_cli_module(cls.__module__)
        cls._description = description
        cls._help = help_
        cls._aliases = defaultdict(set)  # {child: {aliases}}

    def __call__(self, args=None):
        """Call the main functionality of this ViPErLEEDCLI.

        This method must be overridden in concrete subclasses that do
        not have any children, i.e., those that perform any action.
        The overridden method must execute the primary functionality
        of the CLI utility. On the other hand, CLIs with children do
        not need (and typically must not) modify this method to ensure
        proper resolution of the children's functions.

        Parameters
        ----------
        args : Sequence or argparse.Namespace or None, optional
            Command-line arguments for this ViPErLEEDCLI. When
            overriding __call__, use self.parse_cli_args to properly
            resolve both global and specific arguments. If not given
            or None, sys.argv is used instead, as implemented in
            argparse.ArgumentParser. Default is None.

        Returns
        -------
        exit_code : int
            Commonly the following meaning for exit codes is assumed:
            0   -> success
            1   -> clean exit (e.g., via KeyboardInterrupt)
            > 1 -> error
        """
        args = self.parse_cli_args(args)
        try:
            command = args.func
        except AttributeError:  # Called without arguments
            self.parser.parse_args(['--help'])
            return 0  # Unreachable, as help does sys.exit
        return command(args)

    @property
    def children(self):
        """Return all the children of this utility."""
        return self._children

    @property
    def cli_name(self):
        """Return the name of this CLI utility."""
        return self._cli_name

    @property
    def help(self):
        """Return the short help text for this CLI. Used by parents."""
        return self._help

    @property
    def parser(self):
        """Return an ArgumentParser for this CLI with all arguments added."""
        parser = argparse.ArgumentParser(prog=self.cli_name,
                                         description=self._description)
        self.add_parser_arguments(parser)
        return parser

    @classmethod
    def find_sub_utilities(cls):
        """Look up the file system for children of this CLI/utility.

        Returns
        -------
        sub_utilities : tuple
            Fully qualified names of the packages or modules that are
            (potentially) sub-utilities of this one. Packages appear
            before modules. Only those packages that have a cli.py
            module are returned. No check is performed on whether the
            modules (or the cli.py modules of packages) do subclass
            `ViPErLEEDCLI`.
        """
        module_name = cls.__module__
        if not module_name.endswith(f'.{CLI_MODULE_NAME}'):
            # No children for a module-only utility
            return tuple()
        base_path = Path(importlib.util.find_spec(module_name).origin).parent
        all_modules = [module
                       # str(Path) due to cpython/issues/88227
                       for module in pkgutil.iter_modules((str(base_path),))
                       if module.name not in {'__main__', CLI_MODULE_NAME}]
        # All the modules are potentially OK
        modules = (module.name
                   for module in all_modules if not module.ispkg)
        # The packages, however, must have a cli.py module
        packages = []
        for package in all_modules:
            if not package.ispkg:
                continue
            package_path = Path(package.module_finder.path) / package.name
            sub_modules = pkgutil.iter_modules((str(package_path),))
            try:
                cli_module = next(f'{package.name}.{m.name}'
                                  for m in sub_modules
                                  if m.name == CLI_MODULE_NAME)
            except StopIteration:
                continue  # No cli module
            packages.append(cli_module)
        base_name = strip_cli_module(module_name)
        return tuple(f'{base_name}.{module_name}'
                     for module_name in (*packages, *modules))

    @classmethod
    def get_logger(cls):
        """Return a logging.Logger for the correct module/package.

        This method should be used especially by base sub-classes
        of ViPErLEEDCLI that would like to access the same logger
        as the one returned by logging.getLogger(__name__), where
        __name__ is the module that implements the concrete subclass
        of the base class.

        Returns
        -------
        logger: logging.Logger
            A logger for the module in which `cls` is defined
            (except a trailing .cli for package utilities).
        """
        return logging.getLogger(strip_cli_module(cls.__module__))

    @classmethod
    def register_child(cls, child):
        """Register a sub-CLI of this one.

        Parameters
        ----------
        child : str or ViPErLEEDCLI or type
            If a string, it is interpreted as the name of a module that
            defines a ViPErLEEDCLI subclass. Otherwise an instance of a
            ViPErLEEDCLI subclass or a ViPErLEEDCLI subclass itself are
            also acceptable.

        Raises
        ------
        ValueError
            If `child` is a string but errors occur while trying to
            import it as a module.
        NotAViPErLEEDCLIError
            If `child` is a module name but the module does not define
            a concrete `ViPErLEEDCLI` subclass, i.e., one whose name
            does not start with an underscore.
        NotAViPErLEEDCLIError
            If `child` is neither a `ViPErLEEDCLI` subclass nor an
            instance of a `ViPErLEEDCLI` subclass.
        """
        if isinstance(child, str):  # Assume it's a module's name
            child = cls._child_class_from_module_name(child)
        if inspect.isclass(child) and issubclass(child, ViPErLEEDCLI):
            child_cls = child
        elif isinstance(child, ViPErLEEDCLI):
            child_cls = type(child)
        else:
            raise NotAViPErLEEDCLIError('child must be a ViPErLEEDCLI')
        child_name = strip_cli_module(child_cls.__module__)
        cls._children[child_name] = child_cls

    @staticmethod
    def _child_class_from_module_name(module_name):
        """Return a concrete ViPErLEEDCLI subclass from `module_name`."""
        try:
            module = importlib.import_module(module_name)
        except ImportError as exc:
            raise ValueError(f'Cannot register {module_name}.') from exc
        for name, member in inspect.getmembers(module):
            if not inspect.isclass(member):
                continue
            if member in (ViPErLEEDCLI, ViPErLEEDCLIWithAutoChildren):
                continue
            if name.startswith('_'):
                continue
            if issubclass(member, ViPErLEEDCLI):
                return member
        raise NotAViPErLEEDCLIError(f'No ViPErLEEDCLI in module {module_name}')

    @classmethod
    def register_valid_children(cls, children):
        """Register the `children` that subclass ViPErLEEDCLI."""
        for child in children:
            try:
                cls.register_child(child)
            except NotAViPErLEEDCLIError:
                pass

    @classmethod
    def run_as_script(cls):
        """Run this CLI as stand-alone script, reading argv.

        This class method is intended to be used in __main__.py
        files or in "if __name__ == '__main__':"-fenced blocks
        in scripts. It is a safe 'main' function that gracefully
        handles KeyboardInterrupt.

        Raises
        ------
        SystemExit
            With a suitable exit code. If unhandled, this will
            TERMINATE THE INTERPRETER.
        """
        cli = cls()
        try:
            exit_code = cli()
        except KeyboardInterrupt:
            print('Terminated by keyboard interrupt')
            exit_code = 1
        sys.exit(exit_code)

    def add_child_aliases(self, child_cli_name, *aliases):
        """Add `aliases` for `child_cli_name`."""
        self._aliases[child_cli_name].update(aliases)

    def add_global_arguments(self, parser):
        """Add arguments to `parser`. This method is called on all children."""
        parser.add_argument('--version',
                            help='print version number',
                            action='version',
                            version=GLOBALS['version_message'])

    def add_parser_arguments(self, parser):
        """Add CLI arguments for this utility to `parser`.

        This method must be **extended** in subclasses that want to
        introduce new arguments: call super().add_parser_arguments
        in the extended method.

        The base implementation does `add_global_arguments` to parser,
        and adds one sub-parser for each of the children utilities of
        this one. All the added parsers have global arguments as well
        as their own arguments (via their own `add_parser_arguments`).
        This ensures that children utilities are added recursively.

        Arguments
        ---------
        parser : ArgumentParser
            The parser to which new arguments should be added.

        Returns
        -------
        None.
        """
        self.add_global_arguments(parser)
        if not self.children:
            return
        subparsers = parser.add_subparsers(dest='command')
        for child_cls in self.children.values():
            child = child_cls()
            help_txt = child.help or f'call {child.cli_name}'
            aliases = sorted(self._aliases[child.cli_name])
            child_parser = subparsers.add_parser(child.cli_name,
                                                 help=help_txt,
                                                 aliases=aliases)
            child_parser.set_defaults(func=child.__call__)
            # Explicitly add all the child's arguments to the parser
            # created by subparser.add_parser. We would normally do
            # this via child.parser, but we have to do it explicitly
            # here in order to get all the correct sub-children already
            # at this stage. This is important to obtain a proper help
            # message when calling the child without arguments or with
            # the usual -h/--help
            child.add_parser_arguments(child_parser)

    @staticmethod
    def add_verbose_option(parser):
        """Add --verbose flag to `parser`."""
        parser.add_argument('-v', '--verbose',
                            help='increase output verbosity',
                            action='store_true')

    def parse_cli_args(self, args):
        """Return a version of `args` parsed with this CLI's parser.

        Subclasses that need to perform some specific checking of
        arguments can proceed in two ways:
        1. Arguments that do not depend on other arguments should
            be given a dedicated checker in add_parser_arguments
            using the functionality provided by argparse. Examples
            include:
            - parser.add_mutually_exclusive_group;
            - the `type` kwarg of parser.add_argument using a dedicated
                callable, like those implemented in this module (e.g.,
                float_in_zero_one, positive_float)
            - the `action` kwarg of parser.add_argument also using a
                dedicated action, like those implemented in this
                module (e.g., length_choices, required_length,
                minimum_length, maximum_length). Notice that actions
                should use parser.error(msg) to properly report error
                conditions rather than raising ArgumentTypeError.
        2. Arguments that require consistency checks with other parsed
            arguments should be checked in an **extended** version of
            parse_cli_args, or, in rare cases, in a reimplementation
            of __call__ (if some non-CLI-related processing is needed
            for some of the arguments before). These modified methods
            must use self.parser.error(msg) to properly report argument
            errors in the format of argparse rather than raising
            exceptions.

        Parameters
        ----------
        args : Sequence or argparse.Namespace or None
            The command-line arguments to be parsed using self.parser.
            If None, sys.argv is used, as per argparse.ArgumentParser
            implementation.

        Returns
        -------
        parsed_args : argparse.Namespace
        """
        if isinstance(args, argparse.Namespace):
            return args  # Parsed already by someone else
        return self.parser.parse_args(args)


class ViPErLEEDCLIWithAutoChildren(ViPErLEEDCLI):
    """A command-line interface that automatically adds all children."""

    def __init_subclass__(cls, *_, **kwargs):
        """Register all valid sub-utilities found."""
        super().__init_subclass__(**kwargs)
        cls.register_valid_children(cls.find_sub_utilities())



"""ViPErLEED utility: Enforce Symmetry."""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from argparse import ArgumentTypeError

from viperleed.calc import symmetry
from viperleed.guilib.base import PlaneGroup
from viperleed.utilities.poscar.base import _PoscarSymmetryCLI


def validate_planegroup(plane_group):
    """Raise ArgumentTypeError if plane_group is not acceptable."""
    all_groups = PlaneGroup.allGroups.keys()
    if plane_group not in all_groups:
        raise ArgumentTypeError(f'Invalid planegroup: {plane_group}')
    return plane_group


class SymmetrizeSlabCLI(_PoscarSymmetryCLI, cli_name='enforce_symmetry'):
    """Symmetrize a slab according to an optionally specified plane group."""

    long_name = 'symmetrize slab'

    def add_parser_arguments(self, parser):
        """Add an optional --plane-group argument."""
        super().add_parser_arguments(parser)
        parser.add_argument(
            '-g', '--plane-group',
            help=('Plane group to enforce. Default: detected automatically '
                  'from the slab. Use this option to override the automatic '
                  'detection and manually lower the symmetry.'),
            type=validate_planegroup,
            )

    def process_slab(self, slab, args):
        """Find slab symmetry and symmetrize it according to `args`."""
        rpars = self.prepare_rpars(slab, args)
        found_symmetry = symmetry.findSymmetry(slab, rpars)
        plane_group = args.plane_group or found_symmetry
        symmetry.enforceSymmetry(slab, rpars, plane_group)
        return slab


if __name__ == '__main__':
    SymmetrizeSlabCLI.run_as_script()



"""Module base of viperleed.utilities.poscar.

Defines base CLI classes and functions useful to limit code
repetition in poscar utilities.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-27'
__license__ = 'GPLv3+'

from abc import ABC, abstractmethod
from argparse import FileType
from copy import deepcopy
import logging
import sys

from viperleed.calc.classes.rparams import Rparams, SymmetryEps
from viperleed.calc.classes.rparams.defaults import DEFAULTS as PARAM_DEFAULTS
from viperleed.calc.files import poscar
from viperleed.cli_base import ViPErLEEDCLI, positive_float


# ##########################   IMPORTANT  #############################
# All new base classes should be named with a leading underscore to   #
# prevent collection of this module as a valid poscar utility!        #
# #####################################################################


class _PoscarStreamCLI(ViPErLEEDCLI, ABC, cli_name=None):
    """A POSCAR utility that reads/writes from/to files or the terminal."""

    long_name = ''

    def __call__(self, args=None):
        """Call this utility."""
        logger = self.get_logger()
        parsed_args = self.parse_cli_args(args)
        if parsed_args.verbose:
            logger.setLevel(logging.DEBUG)
        logger.info(f'ViPErLEED POSCAR utility: {self.long_name}\n')
        slab = self.read_poscar(parsed_args)
        processed_slab = self.process_slab(slab, parsed_args)
        self.write_output(processed_slab, parsed_args)
        return 0

    def add_infile_argument(self, parser):
        """Add an optional --infile/-i argument to parser.

        This method is automatically called by the base implementation
        of add_parser_arguments. It can be used by subclasses that
        intend to override (and not extend) add_parser_arguments but
        still desire to provide an --infile/-i argument.

        **IMPORTANT**: Subclasses that override add_parser_arguments
        but do not override read_poscar **must** call this method
        in add_parser_arguments.

        Parameters
        ----------
        parser : argparse.ArgumentParser
            The parser to which the optional argument should be
            added. The added argument defaults to the standard-in
            stream, i.e., the terminal.

        Returns
        -------
        None.
        """
        help_ = ('Name of the POSCAR input file. Default: read text '
                 'from the standard-input stream (i.e., the terminal)')
        parser.add_argument('--infile', '-i',
                            type=FileType('r'),
                            help=help_,
                            default=sys.stdin)

    def add_outfile_argument(self, parser, help_=None):
        """Add an optional --outfile/-o argument to parser.

        This method is automatically called by the base implementation
        of add_parser_arguments. It can be used by subclasses that
        intend to override (and not extend) add_parser_arguments but
        still desire to provide an --outfile/-o argument.

        ''IMPORTANT**: Subclasses that override add_parser_arguments
        but do not override write_output **must** call this method
        in add_parser_arguments.

        Parameters
        ----------
        parser : argparse.ArgumentParser
            The parser to which the optional argument should be
            added. The added argument defaults to the standard-out
            stream, i.e., the terminal.

        Returns
        -------
        None.
        """
        help_ = ('Name of the POSCAR output file. Default: write text '
                 'to the standard-output stream (i.e., the terminal)')
        parser.add_argument('--outfile', '-o',
                            type=FileType('w'),
                            help=help_,
                            default=sys.stdout)

    def add_parser_arguments(self, parser):
        """Add generic arguments to this CLI.

        The base implementation adds:
        - all the arguments of ancestor classes.
        - two optional arguments for the input and output files
          (--infile/-i and --outfile/-o, respectively) which
          default to standard-in and standard-out.
        - an optional argument to turn logging verbose.

        **IMPORTANT**: Subclasses may extend this method via super()
        if they want to inherit the arguments of this base class. If
        not, they **must** provide outfile (positional) or --outfile
        (optional) CLI arguments (typically with sys.stdout default),
        unless they also override write_output. If, additionally, they
        do not override read_poscar, they must provide infile/--infile
        CLI arguments (typically with sys.stdin default). Subclasses
        may use the add_infile_argument/add_outfile_argument methods
        to add optional --infile/--outfile arguments.

        Parameters
        ----------
        parser : argparse.ArgumentParser
            The parser to which arguments are added.

        Returns
        -------
        None.
        """
        super().add_parser_arguments(parser)
        self.add_infile_argument(parser)
        self.add_outfile_argument(parser)
        self.add_verbose_option(parser)

    @abstractmethod
    def process_slab(self, slab, args):
        """Return a processed slab from slab.

        This is an abstract method that must be
        extended in concrete subclasses.

        Parameters
        ----------
        slab : Slab
            The slab that was read from stdin and should be processed
            by this utility.
        args : argparse.Namespace
            The parsed command-line arguments.

        Returns
        -------
        processed_slab : Slab
            A new slab processed by this CLI utility. Will be
            written to stout exactly as it is returned.
        """

    def read_poscar(self, args):
        """Return a slab from args.

        The return value of this method is passed on, together with
        args, to self.process_slab. The default implementation reads
        a slab from args.infile, which defaults to the terminal.

        Parameters
        ----------
        args : argparse.Namespace
            The processed CLI arguments.

        Returns
        -------
        slab : Slab
            A slab read using information in args.

        Raises
        ------
        AttributeError
            If subclasses have overridden (and not extended) the
            add_parser_arguments method, have not overridden this
            method and forgot to call add_infile_argument in their
            modified add_parser_arguments.
        """
        try:
            _ = args.infile
        except AttributeError:
            raise AttributeError('--infile argument is missing. If you have '
                                 'overridden add_parser_arguments without '
                                 'also overriding read_poscar, make sure to '
                                 'call add_infile_argument in your overridden '
                                 'add_parser_arguments') from None
        if args.infile.isatty():
            print('Please input the contents of a POSCAR file:')
        try:
            return poscar.read(args.infile)
        except (ValueError, poscar.POSCARError) as exc:
            self.parser.error(f'Failed to read POSCAR. Stopping. Info: {exc}')

    def write_output(self, processed_slab, args):
        """Write output to an output file or the terminal.

        This is the last method that is called by this utility
        before terminating. The default implementation writes
        processed_slab to args.outfile, which defaults to the
        terminal.

        Parameters
        ----------
        processed_slab : Slab
            The slab processed by this utility.
        args : argparse.Namespace
            The processed command-line arguments.
        
        Raises
        ------
        AttributeError
            If subclasses have overridden (and not extended) the
            add_parser_arguments method, have not overridden this
            method and forgot to call add_outfile_argument in their
            modified add_parser_arguments.
        """
        try:
            _ = args.outfile
        except AttributeError:
            raise AttributeError('--outfile argument is missing. If you have '
                                 'overridden add_parser_arguments without '
                                 'also overriding write_output, make sure to '
                                 'call add_outfile_argument in your overridden'
                                 ' add_parser_arguments') from None
        log_level = self.get_logger().getEffectiveLevel()
        poscar.write(processed_slab,
                     filename=args.outfile,
                     comments='none',
                     silent=log_level<=logging.DEBUG)


EPS_DEFAULT = PARAM_DEFAULTS['SYMMETRY_EPS']

class _PoscarSymmetryCLI(_PoscarStreamCLI, ABC, cli_name=None):
    """Base CLI class for utilities dealing with slab symmetry."""

    def add_parser_arguments(self, parser):
        """Add SYMMETRY_EPS optional arguments."""
        parser.add_argument(
            '-e', '--symmetry-eps',
            help=('Epsilon for symmetry detection in angstrom. '
                  f'Default: {EPS_DEFAULT} A'),
            type=positive_float,
            default=EPS_DEFAULT,
            )
        parser.add_argument(
            '--symmetry-eps-z',
            help=('Epsilon for symmetry detection in z in angstrom. If '
                  'not provided, the value of --symmetry-eps is used.'),
            type=positive_float,
            )

    def prepare_rpars(self, slab, args):
        """Return an Rparams with symmetry information, and update slab."""
        rpars = Rparams()
        rpars.SYMMETRY_EPS = SymmetryEps(args.symmetry_eps,
                                         args.symmetry_eps_z)
        rpars.SYMMETRY_FIND_ORI = True
        slab.full_update(rpars)
        return rpars


class _RemoveAtomsCLI(_PoscarStreamCLI, ABC, cli_name=None):
    """A command-line interface that removes atoms from a slab."""

    def add_parser_arguments(self, parser):
        """Add the condition for atom removal."""
        super().add_parser_arguments(parser)
        self.add_remove_condition_arguments(parser)

    @abstractmethod
    def add_remove_condition_arguments(self, parser):
        """Add arguments that define the condition for atom removal."""

    def process_slab(self, slab, args):
        """Return a new slab with atoms from slab removed according to args."""
        modified_slab = deepcopy(slab)
        atoms = self.select_surviving_atoms(modified_slab, args)
        modified_slab.atlist.clear()
        modified_slab.atlist.extend(atoms)
        modified_slab.update_element_count()
        return modified_slab

    @abstractmethod
    def select_surviving_atoms(self, slab, args):
        """Return the atoms of slab that should survive the removal."""



"""Module initialization of viperleed.calc.sections.

ViPErLEED calculation section INITIALIZATION.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-11'
__license__ = 'GPLv3+'

import copy
import logging
import os
from pathlib import Path
import shutil
from zipfile import ZipFile

import numpy as np

from viperleed.calc import ORIGINAL_INPUTS_DIR_NAME
from viperleed.calc import symmetry
from viperleed.calc.classes.rparams import DomainParameters
from viperleed.calc.classes.slab import AlreadyMinimalError
from viperleed.calc.classes.slab import BulkSlab
from viperleed.calc.classes.slab import NoBulkRepeatError
from viperleed.calc.classes.slab import NoVacuumError
from viperleed.calc.classes.slab import Slab
from viperleed.calc.classes.slab import VacuumError
from viperleed.calc.classes.slab import WrongVacuumPositionError
from viperleed.calc.files import beams as iobeams
from viperleed.calc.files import iotensors
from viperleed.calc.files import parameters
from viperleed.calc.files import patterninfo
from viperleed.calc.files import phaseshifts
from viperleed.calc.files import poscar
from viperleed.calc.files import vibrocc
from viperleed.calc.files.beamgen import calc_and_write_beamlist
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.base import angle, rotation_matrix
from viperleed.calc.lib.base import NonIntegerMatrixError
from viperleed.calc.lib.version import Version
from viperleed.calc.lib.woods_notation import writeWoodsNotation
from viperleed.calc.psgen import runPhaseshiftGen, runPhaseshiftGen_old
from viperleed.calc.sections.calc_section import ALL_INPUT_FILES
from viperleed.calc.sections.calc_section import EXPBEAMS_NAMES

logger = logging.getLogger(__name__)


def initialization(sl, rp, subdomain=False):
    """Runs the initialization."""
    if not subdomain:
        rp.try_loading_expbeams_file()
    rp.initTheoEnergies()  # may be initialized based on exp. beams

    if (rp.DOMAINS or rp.domainParams) and not subdomain:
        init_domains(rp)
        return

    _check_slab_duplicates_and_vacuum(sl, rp)

    # if necessary, run findSymmetry:
    if sl.planegroup == "unknown":
        symmetry.findSymmetry(sl, rp)
        symmetry.enforceSymmetry(sl, rp)

    # check whether the slab unit cell is minimal:
    try:
        mincell = sl.get_minimal_ab_cell(rp.SYMMETRY_EPS, rp.SYMMETRY_EPS.z)
    except AlreadyMinimalError:
        transform = np.identity(2)
        reducible = False
    else:
        transform = np.dot(sl.ab_cell.T, np.linalg.inv(mincell)).round()
        reducible = True
    ws = writeWoodsNotation(transform)
    if ws:
        ws = f"= {ws}"
    else:
        ws = "M = {} {}, {} {}".format(*transform.astype(int).ravel())
    if reducible and np.isclose(rp.SYMMETRY_CELL_TRANSFORM,
                                np.identity(2)).all():
        ssl = sl.make_subcell(rp, transform=transform)
        if subdomain:
            rp.SYMMETRY_CELL_TRANSFORM = transform
            logger.info(f"Found SYMMETRY_CELL_TRANSFORM {ws}")
            sl.symbaseslab = ssl
            parameters.modify(rp, "SYMMETRY_CELL_TRANSFORM")                    #TODO: there should probably a comment in the PARAMETERS file?
        else:
            logger.warning(
                f"POSCAR unit cell is not minimal (supercell {ws}). "
                "A minimal POSCAR will be written as POSCAR_mincell for "
                "information. Consider calculating with POSCAR_mincell, "
                "or setting SYMMETRY_CELL_TRANSFORM to conserve "
                "translational symmetry."
                )
            poscar.write(ssl, filename='POSCAR_mincell')
            rp.setHaltingLevel(1)
    elif not np.isclose(rp.SYMMETRY_CELL_TRANSFORM, np.identity(2)).all():
        if not np.isclose(rp.SYMMETRY_CELL_TRANSFORM, transform).all():
            logger.warning("SYMMETRY_CELL_TRANSFORM parameter differs from "
                           f"automatically detected supercell {ws}")
            rp.setHaltingLevel(1)
        if sl.symbaseslab is None:
            try:
                sl.symbaseslab = sl.make_subcell(rp,
                                                 rp.SYMMETRY_CELL_TRANSFORM)
            except ValueError as exc:
                logger.error(f'Invalid SYMMETRY_CELL_TRANSFORM. {exc}')
                raise exc
    if sl.symbaseslab is not None:
        logger.info("A symmetry cell transformation was found. Re-running "
                    "slab symmetry search using base unit cell...")
        symmetry.getSymBaseSymmetry(sl, rp)
        try:
            poscar.write(sl.symbaseslab, filename='POSCAR_mincell',
                         comments='all')
        except OSError:
            logger.warning("Exception occurred while writing POSCAR_mincell")

    # generate new POSCAR
    tmpslab = copy.deepcopy(sl)
    tmpslab.sort_original()
    try:
        poscar.write(tmpslab, filename='POSCAR', comments='all')
    except Exception:
        logger.error("Exception occurred while writing new POSCAR")
        raise
    rp.manifest.append('POSCAR')
    # generate POSCAR_oricell
    tmpslab.revert_unit_cell()
    try:
        poscar.write(tmpslab, filename='POSCAR_oricell', comments='nodir')
    except OSError:
        logger.error("Exception occurred while writing POSCAR_oricell, "
                     "execution will continue...")

    if rp.BULK_LIKE_BELOW > 0:
        if rp.BULK_REPEAT is not None:
            logger.warning('Both BULK_LIKE_BELOW and BULK_REPEAT are defined.'
                           'BULK_LIKE_BELOW will be ignored in favour of the '
                           'explicitly defined bulk repeat vector.')
        else:
            # If successful, the next call updates the relevant rpars
            # attributes: SUPERLATTICE (not written to file),
            # BULK_REPEAT, LAYER_CUTS, and N_BULK_LAYERS (written
            # to file below). The slab will have .layers and the
            # freshly detected .bulkslab.
            sl.detect_bulk(rp)
            vec_str = parameters.modify(
                rp, 'BULK_REPEAT',
                comment='Automatically detected repeat vector'
                )
            parameters.modify(rp, 'LAYER_CUTS')
            parameters.modify(rp, 'N_BULK_LAYERS')
            logger.info(f'Detected bulk repeat vector: {vec_str}')
        parameters.comment_out(rp, 'BULK_LIKE_BELOW')

    # create bulk slab:
    if sl.bulkslab is None:
        sl.make_bulk_slab(rp)
    bsl = sl.bulkslab

    # try identifying bulk repeat:
    if rp.BULK_REPEAT is None:
        try:
            rp.BULK_REPEAT = sl.identify_bulk_repeat(rp.SYMMETRY_EPS,
                                                     rp.SYMMETRY_EPS.z)
        except NoBulkRepeatError:
            pass
        else:
            vec_str = parameters.modify(
                rp, 'BULK_REPEAT',
                comment='Automatically detected repeat vector'
                )
            logger.info(f'Detected bulk repeat vector: {vec_str}')
            # update bulk slab vector
            sl.bulkslab.update_cartesian_from_fractional()
            sl.bulkslab.c_vector[:] = rp.BULK_REPEAT
            sl.bulkslab.collapse_cartesian_coordinates()
    if rp.BULK_REPEAT is None:
        # Failed to detect repeat vector, use fixed distance instead.
        # Assume that interlayer vector from bottom non-bulk to top
        # bulk layer is the same as between bulk units. Save
        # BULK_REPEAT value for later runs, in case atom above moves
        rp.BULK_REPEAT = sl.get_bulk_repeat(rp, only_z_distance=True)
        parameters.modify(
            rp, 'BULK_REPEAT',
            comment='Automatically detected spacing. Check POSCAR_bulk.'
            )
        logger.warning(
            'The BULK_REPEAT parameter was undefined, which may lead to '
            'unintended changes in the bulk unit cell during optimization if '
            'the lowest non-bulk atom moves.\n'
            '# Automatic detection of a bulk repeat vector failed, possibly '
            'because not enough bulk-like layers were found.\n'
            '# The BULK_REPEAT vector is assumed to be parallel to the POSCAR '
            'c vector. Check POSCAR_bulk and the BULK_REPEAT thickness '
            'written to the PARAMETERS file.'
            )
        rp.checklist.append('Check bulk repeat vector in PARAMETERS')
        rp.setHaltingLevel(2)

    if bsl.planegroup == "unknown":
        # find minimum in-plane unit cell for bulk:
        logger.info('Checking bulk unit cell...')
        try:
            sl.ensure_minimal_bulk_ab_cell(rp, warn_convention=True)
        except NonIntegerMatrixError as exc:
            logger.error(
                f'{exc}\n# User-defined SUPERLATTICE will be used instead.'
                )
            rp.setHaltingLevel(2)
        except parameters.errors.InconsistentParameterError as exc:
            # SUPERLATTICE does not match auto-detected one
            logger.warning(exc)
            rp.setHaltingLevel(2)

        if not rp.superlattice_defined:
            ws = writeWoodsNotation(rp.SUPERLATTICE)                   # TODO: replace writeWoodsNotation with guilib functions
            superlattice = rp.SUPERLATTICE.astype(int)
            if ws:
                info = f"= {ws}"
            else:
                info = "M = {} {}, {} {}".format(*superlattice.ravel())
            logger.info(f"Found SUPERLATTICE {info}")

        # bulk plane group detection:
        logger.info("Starting bulk symmetry search...")
        symmetry.findSymmetry(bsl, rp, bulk=True, output=False)
        bsl.revert_unit_cell()  # keep origin matched with main slab
        logger.info(f"Found bulk plane group: {bsl.foundplanegroup}")
        symmetry.findBulkSymmetry(bsl, rp)

        # write POSCAR_bulk
        bsl = copy.deepcopy(sl.bulkslab)
        bsl.sort_original()
        try:
            poscar.write(bsl, filename='POSCAR_bulk', comments='bulk')
        except Exception:
            logger.error("Exception occurred while writing POSCAR_bulk")
            raise

    # write POSCAR_bulk_appended
    n_cells = 1
    if bsl.n_sublayers <= len(bsl.elements)*2:
        # For bulk slabs with very few layers, add a few more
        # repetitions to help the user see better the bulk part
        n_cells += 1
        if bsl.n_sublayers <= len(bsl.elements):
            n_cells += 1
    try:
        poscar.write(sl.with_extra_bulk_units(rp, n_cells)[0],
                     filename='POSCAR_bulk_appended')
    except OSError:
        logger.warning('Exception occurred while writing POSCAR_bulk_appended')

    # Check for an ambiguous angle phi
    _check_and_warn_ambiguous_phi(sl, rp, angle_eps=0.1)

    # Check that layer cuts are not too close together
    _check_and_warn_layer_cuts(rp, sl)

    # check whether PHASESHIFTS are present & consistent:
    newpsGen, newpsWrite = True, True
    # True: new phaseshifts need to be generated/written
    if os.path.isfile("PHASESHIFTS") or os.path.isfile("_PHASESHIFTS"):
        try:
            (rp.phaseshifts_firstline,
             rp.phaseshifts,
             newpsGen,
             newpsWrite) = phaseshifts.readPHASESHIFTS(sl, rp,
                                                       ignoreEnRange=subdomain)
        except Exception:
            logger.warning(
                "Found a PHASESHIFTS file but could not "
                "read it. A new PHASESHIFTS file will be generated."
                "The exception during read was: ", exc_info=True)
            rp.setHaltingLevel(1)
    if rp.TL_VERSION >= Version('1.7.1') and rp.V0_REAL != 'default':
        # check V0_REAL - may have to replace firstline
        if type(rp.V0_REAL) != list:
            logger.warning("Parameter V0_REAL currently does not support "
                           "values other than Rundgren-type potentials."
                           "Input for V0_REAL will be ignored.")
            rp.setHaltingLevel(2)
            newpsGen, newpsWrite = True, True
        else:
            llist = rp.phaseshifts_firstline.split()
            try:
                flist = [float(llist[i+1]) for i in range(4)]
            except (ValueError, IndexError):
                firstline_write = True
            else:
                firstline_write = any(abs(flist[i] - rp.V0_REAL[i]) >= 1e-2
                                      for i in range(4))
            if firstline_write:
                newpsWrite = True
                rp.phaseshifts_firstline = (
                    rp.phaseshifts_firstline[:4]
                    + "{:8.2f}{:8.2f}{:8.2f}{:8.2f}".format(*rp.V0_REAL)
                    + rp.phaseshifts_firstline[36:]
                    )
    if newpsGen:
        # Check for old executable. Used to be called EEASiSSS.x
        if (not Path(rp.source_dir / "eeasisss").is_file()
                and Path(rp.source_dir / "EEASiSSS.x").is_file()):
            rundgrenpath = 'EEASiSSS.x'
        else:
            # let psgen catch the error if neither executable is found
            rundgrenpath = 'eeasisss'
        serneliuspath = 'seSernelius'
        atom_density_path = 'atom_density_files'
        logger.info("Generating phaseshifts data... ")
        ps_gen, kwargs = runPhaseshiftGen, {}
        if rp.PHASESHIFTS_CALC_OLD:
            ps_gen = runPhaseshiftGen_old
            kwargs = {"excosource": serneliuspath,
                      "atdenssource": atom_density_path}
        try:
            (rp.phaseshifts_firstline,
             rp.phaseshifts) = ps_gen(sl, rp, psgensource=rundgrenpath,
                                      **kwargs)
        except Exception:
            logger.error("Exception while calling phaseshiftgen: ")
            raise
        else:
            logger.debug("Finished generating phaseshift data")
    if newpsWrite:
        try:
            phaseshifts.writePHASESHIFTS(rp.phaseshifts_firstline,
                                         rp.phaseshifts)
        except Exception:
            logger.error("Exception during writePHASESHIFTS: ")
            raise
    rp.fileLoaded["PHASESHIFTS"] = True
    rp.updateDerivedParams()
    rp.manifest.append("PHASESHIFTS")
    try:
        phaseshifts.plot_phaseshifts(sl, rp)
    except Exception:
        logger.warning("Failed to plot phaseshifts", exc_info=rp.is_debug_mode)

    # generate beamlist
    logger.info("Generating BEAMLIST...")
    calc_and_write_beamlist(sl, rp, beamlist_name="BEAMLIST")

    try:
        rp.beamlist = iobeams.readBEAMLIST()
        rp.fileLoaded["BEAMLIST"] = True
    except Exception:
        logger.error("Error while reading required file BEAMLIST")
        raise

    if not subdomain:
        patterninfo.writePatternInfo(sl, rp)

        # if EXPBEAMS was loaded, it hasn't been checked yet - check now
        if rp.fileLoaded["EXPBEAMS"]:
            iobeams.checkEXPBEAMS(sl, rp)
        # write and sort IVBEAMS
        if not rp.fileLoaded["IVBEAMS"]:
            try:
                rp.ivbeams = iobeams.writeIVBEAMS(sl, rp)
                rp.ivbeams_sorted = False
                rp.fileLoaded["IVBEAMS"] = True
                rp.manifest.append("IVBEAMS")
            except Exception:
                logger.error("Error while writing IVBEAMS file based on "
                             "EXPBEAMS data.")
                raise
    if rp.fileLoaded["IVBEAMS"] and not rp.ivbeams_sorted:
        rp.ivbeams = iobeams.sortIVBEAMS(sl, rp)
        rp.ivbeams_sorted = True

    # Create directory compile_logs in which logs from compilation will be saved
    make_compile_logs_dir(rp)

    # At the end of initialization preserve the original input files
    _preserve_original_input(rp, logger)
    return


def init_domains(rp):
    """Runs an alternative initialization for the domain search. This will
    include running the 'normal' initialization for each domain."""
    rp.try_loading_expbeams_file()
    rp.initTheoEnergies()  # may be initialized based on exp. beams
    if len(rp.DOMAINS) < 2:
        logger.error("A domain search was defined, but less than two domains "
                     "are defined. Execution will stop.")
        rp.setHaltingLevel(3)
        return
    checkFiles = ["POSCAR", "PARAMETERS", "VIBROCC", "PHASESHIFTS"]
    home = Path.cwd().resolve()
    for name, path in rp.DOMAINS.items():
        # determine the target path
        target = Path(f"Domain_{name}").resolve()
        dp = DomainParameters(target, home, name)
        if target.is_dir():
            logger.warning(f"Folder {target} already exists. "
                           "Contents may get overwritten.")
        else:
            target.mkdir()
        logger.info(f"Fetching input files for domain {name}")
        if os.path.isdir(path):
            # check the path for Tensors
            tensorIndex = leedbase.getMaxTensorIndex(path)
            if tensorIndex != 0:
                try:
                    iotensors.getTensors(tensorIndex, base_dir=path,
                                         target_dir=target)
                except Exception as exc:
                    tensorIndex = 0
                    logger.warning(f"Error fetching Tensors: {exc}")
            if tensorIndex != 0:
                tensorDir = target / "Tensors" / f"Tensors_{tensorIndex:03d}"
                for file in (checkFiles + ["IVBEAMS"]):
                    if os.path.isfile(tensorDir / file):
                        shutil.copy2(tensorDir / file, target / file)
                    else:
                        logger.warning(f"Input file {file} is missing in "
                                       "Tensors directory. A new reference "
                                       "calculation is required.")
                        tensorIndex = 0
                        break
            if tensorIndex != 0:
                dp.tensorDir = tensorDir
            else:       # no usable tensors in that dir; get input
                dp.refcalcRequired = True
                logger.info("No previous Tensors found, reference calculation "
                            "is required.")
                for file in checkFiles:
                    if os.path.isfile(os.path.join(path, file)):
                        try:
                            shutil.copy(os.path.join(path, file),
                                        os.path.join(target, file))
                        except Exception:
                            if file != "PHASESHIFTS":
                                logger.error(
                                    f"Error copying required file {file} for "
                                    f"domain {name} from origin folder {path}"
                                    )
                                raise RuntimeError("Error getting domain "
                                                   "input files")
                    elif file != "PHASESHIFTS":
                        logger.error(f"Required file {file} for domain {name} "
                                     f"not found in origin folder {path}")
                        raise RuntimeError("Error getting domain input files")
        elif os.path.isfile(path):
            try:
                tensorIndex = leedbase.getMaxTensorIndex(target)
            except Exception:
                tensorIndex = 0
            tensorDir = target / "Tensors" / f"Tensors_{tensorIndex + 1:03d}"
            try:
                tensorDir.mkdir(parents=True, exist_ok=True)
            except Exception:
                raise
            try:
                with ZipFile(path, 'r') as archive:
                    archive.extractall(tensorDir)                               # TODO: maybe it would be nicer to read directly from the zip file
            except Exception:
                logger.error("Failed to unpack Tensors for domain "
                             f"{name} from file {path}")
                raise RuntimeError("Error getting domain input files")
            for file in (checkFiles + ["IVBEAMS"]):
                if (tensorDir / file).is_file():
                    shutil.copy2(tensorDir / file, target / file)
                else:
                    logger.error(f"Required file {file} for domain {name} not "
                                 f"found in Tensor directory {tensorDir}")
                    raise RuntimeError("Error getting domain input files")
            dp.tensorDir = tensorDir
        try:
            # initialize for that domain
            os.chdir(target)
            logger.info(f"Reading input files for domain {name}")
            try:
                dp.sl = poscar.read()
                dp.rp = parameters.read()                                       # NB: if we are running from stored Tensors, then these parameters will be stored versions, not current PARAMETERS from Domain directory
                warn_if_slab_has_atoms_in_multiple_c_cells(dp.sl, dp.rp, name)
                dp.rp.workdir = home
                dp.rp.source_dir = rp.source_dir
                dp.rp.timestamp = rp.timestamp
                parameters.interpret(dp.rp, slab=dp.sl,
                                     silent=rp.LOG_LEVEL > logging.DEBUG)
                dp.sl.full_update(dp.rp)
                dp.rp.fileLoaded["POSCAR"] = True
                dp.rp.updateDerivedParams()
                try:
                    vibrocc.readVIBROCC(dp.rp, dp.sl)
                    dp.rp.fileLoaded["VIBROCC"] = True
                except Exception:
                    logger.error("Error while reading required file VIBROCC")
                    raise
                dp.sl.full_update(dp.rp)
                try:
                    dp.rp.ivbeams = iobeams.readIVBEAMS()
                except FileNotFoundError:
                    pass
                except Exception:
                    logger.error(
                        f"Error while reading IVBEAMS for domain {name}"
                        )
                else:
                    dp.rp.ivbeams_sorted = False
                    dp.rp.fileLoaded["IVBEAMS"] = True
            except Exception:
                logger.error("Error loading POSCAR and "
                             f"PARAMETERS for domain {name}")
                raise
            logger.info(f"Running initialization for domain {name}")
            try:
                initialization(dp.sl, dp.rp, subdomain=True)
            except Exception:
                logger.error(f"Error running initialization for domain {name}")
                raise
            rp.domainParams.append(dp)
        except Exception:
            logger.error(f"Error while initializing domain {name}")
            raise
        finally:
            os.chdir(home)
    if len(rp.domainParams) < len(rp.DOMAINS):
        raise RuntimeError("Failed to read domain parameters")
    # check whether bulk unit cells match
    logger.info("Starting domain consistency check...")
    bulkuc0 = rp.domainParams[0].sl.bulkslab.ab_cell.T
    eps = 1e-4
    for dp in rp.domainParams[1:]:
        bulkuc = dp.sl.bulkslab.ab_cell.T
        if np.all(abs(bulkuc-bulkuc0) < eps):
            continue
        # if the unit cells don't match right away, try if rotation matches
        if (all(abs(np.linalg.norm(bulkuc0[i]) - np.linalg.norm(bulkuc[i]))
                < eps for i in range(0, 2))
                and abs(angle(bulkuc[0], bulkuc[1])
                        - angle(bulkuc0[0], bulkuc0[1])) < eps):
            logger.info(f"Bulk unit cells of domain {rp.domainParams[0].name} "
                        f"and domain {dp.name} are mismatched, but can be "
                        f"matched by rotating domain {dp.name}.")
            ang = angle(bulkuc[0], bulkuc0[0])
            dp.sl.apply_matrix_transformation(rotation_matrix(ang, dim=3))      # TODO: this changes the coordinate frame. We need to modify BEAM_INCIDENCE! Issue #69, PR #73
        else:
            logger.error(f"Bulk unit cells of domain {rp.domainParams[0].name}"
                         f" and domain {dp.name} are mismatched, and cannot be"
                         "matched by rotation. Domain search cannot proceed. "
                         "Execution will stop.")
            rp.setHaltingLevel(3)
            return
    logger.debug("Domain bulk unit cells are compatible.")
    uc0 = rp.domainParams[0].sl.ab_cell.T
    largestDomain = rp.domainParams[0]
    allMatched = all(np.all(abs(dp.sl.ab_cell.T - uc0) < 1e-4)
                     for dp in rp.domainParams[1:])
    supercellRequired = []
    if allMatched:
        logger.debug("Domain surface unit cells are matched.")
    else:
        maxArea = abs(np.linalg.det(uc0))
        for dp in rp.domainParams[1:]:
            uc = dp.sl.ab_cell.T
            if abs(np.linalg.det(uc)) > maxArea:
                maxArea = abs(np.linalg.det(uc))
                largestDomain = dp
        uc0 = largestDomain.sl.ab_cell.T
        for dp in [p for p in rp.domainParams if p != largestDomain]:
            uc = dp.sl.ab_cell.T
            if not np.all(abs(uc-uc0) < 1e-4):
                dp.refcalcRequired = True
                trans = np.dot(uc0, np.linalg.inv(uc))
                if np.any(abs(trans - np.round(trans)) > 1e-4):
                    logger.error(
                        f"Surface unit cell of domain {dp.name} cannot be "
                        "transformed to the largest surface unit cell (domain "
                        f"{largestDomain.name}) by an integer transformation. "
                        "Execution will stop. Please supply all domain "
                        "structures as matching supercells."
                        )
                    rp.setHaltingLevel(3)
                    return
                else:
                    supercellRequired.append(dp)
                    oldslab = dp.sl
                    dp.sl = dp.sl.make_supercell(np.round(trans))
                    dp.rp.SUPERLATTICE = largestDomain.rp.SUPERLATTICE.copy()
                    dp.sl.symbaseslab = oldslab
                    dp.rp.SYMMETRY_CELL_TRANSFORM = trans
                    parameters.modify(dp.rp, "SYMMETRY_CELL_TRANSFORM",
                                      path=dp.workdir)
        logger.info("Domain surface unit cells are mismatched, but can be "
                    "matched by integer transformations.")
    # store some information about the supercell in rp:
    rp.pseudoSlab = Slab()
    rp.pseudoSlab.ucell = largestDomain.sl.ucell.copy()
    rp.pseudoSlab.bulkslab = BulkSlab()
    rp.pseudoSlab.bulkslab.ucell = largestDomain.sl.bulkslab.ucell.copy()
    # run beamgen for the whole system
    logger.info("Generating BEAMLIST...")
    calc_and_write_beamlist(copy.deepcopy(largestDomain.sl),
                      rp,
                      domains=True,
                      beamlist_name='BEAMLIST')
    try:
        rp.beamlist = iobeams.readBEAMLIST()
        rp.fileLoaded["BEAMLIST"] = True
    except Exception:
        logger.error("Error while reading required file BEAMLIST")
        raise
    # if EXPBEAMS was loaded, it hasn't been checked yet - check now
    if rp.fileLoaded["EXPBEAMS"]:
        iobeams.checkEXPBEAMS(None, rp, domains=True)
    # write and sort IVBEAMS
    if not rp.fileLoaded["IVBEAMS"]:
        try:
            rp.ivbeams = iobeams.writeIVBEAMS(None, rp, domains=True)
            rp.ivbeams_sorted = False
            rp.fileLoaded["IVBEAMS"] = True
            rp.manifest.append("IVBEAMS")
        except Exception:
            logger.error("Error while writing IVBEAMS file based on "
                         "EXPBEAMS data.")
            raise
    if not rp.ivbeams_sorted:
        rp.ivbeams = iobeams.sortIVBEAMS(None, rp)
        rp.ivbeams_sorted = True

    rp.updateDerivedParams()  # Also sets LMAX
    if not rp.LMAX.has_max:
        rp.LMAX.max = max(dp.rp.LMAX.max for dp in rp.domainParams)
    for dp in rp.domainParams:
        if dp.refcalcRequired:
            continue
        cmessage = f"Reference calculation required for domain {dp.name}: "
        # check energies
        if not dp.rp.THEO_ENERGIES.contains(rp.THEO_ENERGIES):
            logger.info("%sEnergy range is mismatched.", cmessage)
            dp.refcalcRequired = True
            continue
        # check LMAX - should be obsolete since TensErLEED version 1.6
        if rp.TL_VERSION <= Version('1.6.0') and rp.LMAX.max != dp.rp.LMAX.max:
            logger.info("%sLMAX is mismatched.", cmessage)
            dp.refcalcRequired = True
        # check beam incidence
        if rp.THETA != dp.rp.THETA or rp.PHI != dp.rp.PHI:
            logger.info("%sBEAM_INCIDENCE is mismatched.", cmessage)
            dp.refcalcRequired = True
        # check IVBEAMS
        if not dp.rp.fileLoaded["IVBEAMS"]:
            logger.info("%sNo IVBEAMS file loaded", cmessage)
            dp.refcalcRequired = True
            continue
        if (len(rp.ivbeams) != len(dp.rp.ivbeams)
                or not all(dp.rp.ivbeams[i].isEqual(rp.ivbeams[i])
                           for i in range(len(rp.ivbeams)))):
            logger.info("%sIVBEAMS file mismatched with supercell.", cmessage)
            dp.refcalcRequired = True
            continue

    rr = [dp for dp in rp.domainParams if dp.refcalcRequired]
    if rr:
        logger.info("The following domains require new reference "
                    f"calculations: {', '.join(d.name for d in rr)}")
        for dp in rp.domainParams:
            for var in ["THEO_ENERGIES", "THETA", "PHI", "N_CORES", "ivbeams",
                        "source_dir"]:
                setattr(dp.rp, var, copy.deepcopy(getattr(rp, var)))
            if rp.TL_VERSION <= Version('1.6.0'):  # not required since TensErLEED v1.61
                dp.rp.LMAX.max = rp.LMAX.max

    # repeat initialization for all slabs that require a supercell
    for dp in supercellRequired:
        logger.info("Re-running initialization with "
                    f"supercell slab for domain {dp.name}")
        try:
            os.chdir(dp.workdir)
            dp.sl.clear_symmetry_and_ucell_history()
            dp.rp.SYMMETRY_FIND_ORI = True
            initialization(dp.sl, dp.rp, subdomain=True)
        except Exception:
            logger.error(f"Error while re-initializing domain {dp.name}")
            raise
        finally:
            os.chdir(home)

    if 4 not in rp.RUN and 1 not in rp.RUN and rr:
        logger.error(
            "Some domains require new reference calculations before "
            "a domain search can be executed. Please either manually "
            "execute appropriate reference calculations, or set RUN = 4"
            )
        rp.setHaltingLevel(3)
        return

    while 4 in rp.RUN:
        if rr:
            rp.RUN.insert(rp.RUN.index(4), 1)
        rp.RUN.insert(rp.RUN.index(4), 2)
        rp.RUN.insert(rp.RUN.index(4), 3)
        rp.RUN.remove(4)


def _preserve_original_input(rp, init_logger, path=""):
    """Create original_inputs directory and copies the input files there."""
    path = Path(path)
    orig_inputs_path = path / ORIGINAL_INPUTS_DIR_NAME
    try:
        orig_inputs_path.mkdir(parents=True, exist_ok=True)
    except OSError as exc:
        raise RuntimeError(f"Could not create directory "
                           f"{ORIGINAL_INPUTS_DIR_NAME}. "
                           "Check disk permissions.") from exc

    # We will copy all files that have potentially been used as
    # inputs. Make sure the correct version of EXPBEAMS is stored
    files_to_preserve = ALL_INPUT_FILES.copy()
    files_to_preserve.remove('EXPBEAMS')
    if rp.expbeams_file_name:
        files_to_preserve.add(rp.expbeams_file_name)

    # copy all files to orig_inputs that were used as original input
    for file in files_to_preserve:
        # save under name EXPBEAMS.csv
        file_path = path / file
        if not file_path.is_file():
            init_logger.warning(f"Could not find file {file}. "
                                    "It will not be stored in "
                                    f"{ORIGINAL_INPUTS_DIR_NAME}.")
            rp.setHaltingLevel(1)
            return
        try:
            shutil.copy2(file_path, orig_inputs_path)
        except OSError:
            init_logger.warning(f"Could not copy file {file} to "
                                f"{ORIGINAL_INPUTS_DIR_NAME}.")
            rp.setHaltingLevel(1)


def make_compile_logs_dir(rp):
    """Create compile_logs directory where compilation logs are saved."""
    # put into rp
    rp.compile_logs_dir = Path(rp.workdir) / "compile_logs"

    # makes compile_logs directory
    try:
        rp.compile_logs_dir.mkdir(exist_ok=True)
    except OSError:
        logger.warning(f"Could not create directory {rp.compile_logs_dir}")
        rp.setHaltingLevel(1)


def warn_if_slab_has_atoms_in_multiple_c_cells(slab, rpars, domain_name=''):
    """Log a WARNING if slab's atoms do not all belong to the same cell.

    It only makes sense to use this function right after `slab` has
    been loaded (e.g., via poscar.read or .from_ase) and before any
    of the c-collapsing functions are called. These include:
        .collapse_cartesian_coordinates
        .collapse_fractional_coordinates
        .full_update

    Parameters
    ----------
    slab : SurfaceSlab
        The slab to be checked.
    rpars : Rparams
        The current PARAMETERS object. Used for logging purposes only.
    domain_name : str, optional
        The name of the structural domain to which slab belongs.
        Used only for logging purposes. Default is an empty string.

    Returns
    -------
    None.
    """
    _msg = 'POSCAR file ' + f'of domain {domain_name} ' if domain_name else ''
    _msg += ('has some atoms outside the base unit cell along the third unit '
             'vector (i.e., they have fractional c coordinates smaller/larger '
             'than 0/1). These atoms will be back-folded into the base unit '
             'cell assuming periodicity along c. This will impact layer '
             'creation and may modify the thickness of the vacuum gap '
             'detected. Check POSCAR to ensure correct layer assignment.')
    if slab.has_atoms_in_multiple_c_cells():
        logger.warning(_msg)
        rpars.setHaltingLevel(1)


def _check_and_warn_ambiguous_phi(sl, rp, angle_eps=0.1):
    """Check if phi is ambiguous and warn if so."""
    angle_between_first_uc_vec_and_x = sl.angle_between_ucell_and_coord_sys
    if angle_between_first_uc_vec_and_x > angle_eps and rp.THETA > angle_eps:
        logger.info(
            f"Detected non-zero angle theta ({rp.THETA:.2f})° and"
            f"an angle of {angle_between_first_uc_vec_and_x:.2f}° "
            "between the first unit cell vector and the x direction of "
            "the coordinate system in the POSCAR file.\n"
            "Make sure the angle phi is interpreted correctly: "
            f"Phi is {rp.PHI:.2f}° from x, which is "
            f"{(rp.PHI+ angle_between_first_uc_vec_and_x):.2f}° from a.\n"
            "See the ViPErLEED documentation for the parameter BEAM_INCDIDENCE "
            "for details."
            )


def _check_and_warn_layer_cuts(rpars, slab):
    """Check if layer cuts are too close together and warn if so."""
    layer_cuts = rpars.LAYER_CUTS
    min_spacing = slab.smallest_interlayer_gap
    if min_spacing < 1.0:
        logger.warning(
            f"Layer cuts are very close together. The minimum spacing "
            f"between layers is {min_spacing:.2f} Å. This may lead to "
            "covergence issues in the reference calculation. Check the "
            "LAYERS_CUTS parameter in the PARAMETERS file."
            )


def _check_slab_duplicates_and_vacuum(slab, rpars):
    """Complain if slab has duplicate atoms or an inadequate vacuum gap."""
    # Make sure that there are no duplicate atoms
    slab.check_atom_collisions(rpars.SYMMETRY_EPS)

    # Make sure that there's enough vacuum. Stop in complex situations
    try:
        slab.check_vacuum_gap()
    except VacuumError as exc:
        exc.fixed_slab.sort_original()  # Rather than by z
        poscar.write(exc.fixed_slab, 'POSCAR_vacuum_corrected')
        exc_type = type(exc)
        _msg = exc.message
        _msg += '. This may cause problems with layer assignment! '
        _msg += 'You can find a POSCAR_vacuum_corrected file in SUPP with '
        _msg += ('the correct position of vacuum. '
                 if exc_type is WrongVacuumPositionError
                 else 'a large enough vacuum gap. ')
        _msg += (
            'If you intend to use it for a new run, be careful to edit any '
            'PARAMETERS expressed as fractions of the c unit vector (e.g., '
            'LAYERS_CUTS, BULK_LIKE_BELOW, BULK_REPEAT, ...)'
            )
        # Notice the straight type check rather than isinstance, as
        # NoVacuumError is a subclass of NotEnoughVacuumError, which
        # is, instead, the acceptable case in which we warn below
        if exc_type in (NoVacuumError, WrongVacuumPositionError):
            raise exc_type(_msg, exc.fixed_slab) from None
        rpars.setHaltingLevel(1)
        logger.warning(_msg)
    if not slab.layers:
        # May have been cleared by shifting slab away from c==0
        slab.create_layers(rpars)



"""
===============================================
      ViPErLEED Graphical User Interface
===============================================
 *** module guilib.leedsim.ExportCSVDialog ***

Created: 2020-01-11
Author: Michele Riva

"""

from quicktions import Fraction

import numpy as np
import PyQt5.QtCore as qtc
import PyQt5.QtGui as qtg
import PyQt5.QtWidgets as qtw

# import guilib as gl
from viperleed import guilib as gl


class ExportCSVDialog(qtw.QDialog):
    # this signal contains the arguments to be passed on to export_csv
    exportSelected = qtc.pyqtSignal(dict)
    
    
    def __init__(self, leed, parent=None):
        if not isinstance(leed, gl.LEEDPattern):
            raise
        #super(ExportCSVDialog, self).__init__(parent)
        super().__init__(parent)
        self.leed = leed
        self.setWindowModality(qtc.Qt.WindowModal)
        flags = self.windowFlags()
        flags &= ~qtc.Qt.WindowCloseButtonHint  # disable close button
        self.setWindowFlags(flags)
        self.setWindowTitle('Select export data')
        
        self.compose()
        self.connectControls()  # connect some signals
        self.open()
    
    def compose(self):
        font = gl.AllGUIFonts().labelFont
        if self.leed.n_domains > 1:  # more than one domain
            domColors = self.leed.domColors
        else:
            domColors = [(0, 0, 0)]  # black
        
        # textbox for optional name to place in the header
        stuctNameLab = qtw.QLabel('Structure name (optional):')
        self.structName = qtw.QLineEdit('')
        self.structName.setFont(font)
        self.structName.setToolTip('This name will be included in the header'
                                   ' of the exported file')
        
        nameLay = qtw.QVBoxLayout()
        nameLay.addWidget(stuctNameLab)
        nameLay.addWidget(self.structName)
        nameLay.addStretch(1)
        
        txt = qtw.QLabel('Select which domains to export')
        txt.setFont(font)
        
        # Radio buttons to select whether all domains should be exported
        # or only some specific ones
        all = qtw.QRadioButton('All')
        visible = qtw.QRadioButton('Visible domains')
        selection = qtw.QRadioButton('Other')
        self.exportRadio = (all, visible, selection)
        
        for radio in self.exportRadio:
            radio.setFont(font)
        self.exportRadio[0].setChecked(True)
        
        # And 'Export' and 'Cancel' buttons
        self.doneBut = qtw.QPushButton('Export')
        self.cancelBut = qtw.QPushButton('Cancel')
        for but in [self.doneBut, self.cancelBut]:
            but.setFont(font)
            but.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Fixed)
        butsLay = qtw.QHBoxLayout()
        butsLay.addStretch(1)
        butsLay.addWidget(self.doneBut)
        butsLay.addWidget(self.cancelBut)
        
        # Prepare also as many tick boxes as there are domains
        self.domTicks = [qtw.QCheckBox() for dom in range(self.leed.n_domains)]
        for dom, (color, tick) in enumerate(zip(domColors, self.domTicks)):
            tick.setFont(font)
            tick.setText('Dom. %d' % (dom + 1))
            p = tick.palette()  # for coloring text the same as the domains
            p.setColor(qtg.QPalette.WindowText, qtg.QColor.fromRgbF(*color))
            tick.setPalette(p)
            tick.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Preferred)
            tick.hide()
        
        # and arrange them in a QGridLayout, in a similar manner as the 
        # matrices in the MatricesPopup
        nDoms = len(self.domTicks)  # this is 1, 2, 3, 4, 6, 8, or 12
        if nDoms in range(1, 4):  # 1 -- 3
            nRows = 1
        elif nDoms in range(4, 10):  # 4 -- 8
            nRows = 2
        else:
            nRows = 3
        nCols = int(nDoms/nRows)
        
        ticksLay = qtw.QGridLayout()
        [ticksLay.addWidget(tick, index//nCols, index % nCols)
         for (index, tick) in enumerate(self.domTicks)]
        
        # And build the dialog by putting the widgets in a layout
        diagLay = qtw.QVBoxLayout()
        diagLay.addLayout(nameLay)
        diagLay.addWidget(txt)
        [diagLay.addWidget(radio) for radio in self.exportRadio]
        diagLay.addLayout(ticksLay)
        diagLay.addStretch(1)
        diagLay.addLayout(butsLay)
        # diagLay.setSizeConstraint(QLayout.SetFixedSize)
        
        self.setLayout(diagLay)
        
        # In case there is only one domain, there's no reason to show all these
        # controls, except for the optional name and the buttons
        if nDoms == 1:
            children = gl.get_all_children_widgets(diagLay)
            keepVisible = (gl.get_all_children_widgets(nameLay)
                           | gl.get_all_children_widgets(butsLay))
            for child in children - keepVisible:
                child.hide()
    
    def connectControls(self):
        self.exportRadio[2].toggled.connect(self.exportSomeTriggered)
        for but in [self.doneBut, self.cancelBut]:
            but.clicked.connect(self.onButtonPressed)
    
    def exportSomeTriggered(self, checked=None):
        if checked is None:
            return
        
        if checked:
            [tick.show() for tick in self.domTicks]
        else:
            [tick.hide() for tick in self.domTicks]
    
    def onButtonPressed(self, checked):
        btn = self.sender()
        if btn == self.doneBut:
            if self.exportRadio[0].isChecked():
                # export all domains
                [tick.setChecked(True) for tick in self.domTicks]
            elif self.exportRadio[1].isChecked():
                # export visible domains
                # TODO: read in which ones are visible. Probably an attribute 
                # in self.leed
                # for now behaves the same as the previous
                [tick.setChecked(True) for tick in self.domTicks]
            params = self.pack_export_params()
            self.exportSelected.emit(params)
            self.accept()
        elif btn == self.cancelBut:
            self.reject()
    
    def pack_export_params(self):
        """
        Prepares the parameters that define what is to be exported
        """
        params = dict()
        params['domains'] = ([i for (i, tick) in enumerate(self.domTicks)
                             if tick.isChecked()],)
        params['name'] = self.structName.text()
        
        # Other parameters (leed, source, file names) are taken care of
        # by LEED_GUI
        
        return params




"""Cleanup functions, to be used between sections or before/after execution."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2021-06-04'
__license__ = 'GPLv3+'

import logging
import os
from pathlib import Path                                                        # TODO: use everywhere
import re
import shutil
import time
from timeit import default_timer as timer
from zipfile import ZIP_DEFLATED, ZipFile

from viperleed.calc import DEFAULT_WORK_HISTORY
from viperleed.calc import LOG_PREFIX
from viperleed.calc import ORIGINAL_INPUTS_DIR_NAME
from viperleed.calc.lib.base import copytree_exists_ok,get_elapsed_time_str

# files to go in SUPP
_SUPP_FILES = (
    "AUXBEAMS",
    "AUXEXPBEAMS",
    "AUXGEO",
    "AUXLATGEO",
    "AUXNONSTRUCT",
    "BEAMLIST",
    "delta-input",
    "EEASISSS-input.txt",
    "eeasisss-input",
    "EEASISSS-log.txt",
    "muftin.f",
    "Phaseshifts_plots.pdf",
    "POSCAR_bulk",
    "POSCAR_bulk_appended",
    "POSCAR_mincell",
    "POSCAR_oricell",
    "POSCAR_vacuum_corrected",
    "refcalc-FIN",
    "refcalc-PARAM",
    "restrict.f",
    "rfactor-WEXPEL",
    "rfactor-PARAM",
    "search-PARAM",
    "search-rf.info",
    "search.steu",
    "searchpars.info",
    "superpos-CONTRIN",
    "superpos-PARAM",
    )

_SUPP_DIRS = (ORIGINAL_INPUTS_DIR_NAME, "compile_logs")

# files to go in OUT
_OUTFILES = ("THEOBEAMS.csv", "THEOBEAMS_norm.csv", "THEOBEAMS.pdf",
             "PatternInfo.tlm", "SD.TL", "refcalc-fd.out", "refcalc-amp.out",
             "Rfactor_plots_refcalc.pdf", "control.chem",
             "Search-progress.pdf", "Search-progress.csv",
             "Search-report.pdf", "FITBEAMS.csv", "FITBEAMS_norm.csv",
             "superpos-spec.out", "Rfactor_plots_superpos.pdf",
             "Rfactor_analysis_refcalc.pdf",
             "Rfactor_analysis_superpos.pdf",
             "Errors_summary.csv", "Errors.zip", "Errors.pdf",
             "FD_Optimization.csv", "FD_Optimization.pdf",
             "FD_Optimization_beams.pdf", "Complex_amplitudes_imag.csv",
             "Complex_amplitudes_real.csv")

# output files that can be used as input in future runs - keep during prerun
iofiles = ["control.chem", "refcalc-fd.out", "superpos-spec.out"]

logger = logging.getLogger(__name__)


def prerun_clean(rp, logname=""):
    """Clean up the work directory before viperleed.calc starts.

    Delete workhistory, old executables, and old logfiles.
    Call move_oldruns if required.

    Parameters
    ----------
    rp : Rparams
        The run parameters, needed for move_oldruns.
    logname : str, optional
        Name of the current log file, to be excluded from cleanup.

    Returns
    -------
    None.
    """
    # clean out the workhistory folder, if there is one
    if os.path.isdir(os.path.join(".", DEFAULT_WORK_HISTORY)):
        try:
            shutil.rmtree(os.path.join(".", DEFAULT_WORK_HISTORY))
        except Exception:
            logger.warning(f"Failed to clear {DEFAULT_WORK_HISTORY} folder.")
    # get rid of old POSCAR_OUT, VIBROCC_OUT and R_OUT files:
    for d in [".", os.path.join(".", "OUT")]:
        if os.path.isdir(d):
            for s in ["POSCAR_OUT", "VIBROCC_OUT", "R_OUT"]:
                for file in [fn for fn in os.listdir(d) if fn.startswith(s)]:
                    try:
                        os.remove(os.path.join(d, file))
                    except Exception:
                        logger.debug("Failed to delete file {}"
                                     .format(os.path.join(d, file)))
    # clean up old executable files:
    for fn in ["refcalc", "rfactor", "search", "superpos"]:
        p = re.compile(fn+r'-\d{6}-\d{6}')
        for file in [f for f in os.listdir()
                  if len(f) == len(fn) + 14 and p.match(f)]:
            try:
                os.remove(file)
            except Exception:
                logger.debug(f"Failed to delete file {file}")
    # see if there are old logfiles
    old_logs = [f for f in os.listdir() if os.path.isfile(f) and
               f.endswith(".log") and f != logname]
    if len(old_logs) > 0:
        try:
            move_oldruns(rp, prerun=True)
        except Exception:
            logger.warning("Exception while trying to clean up from previous "
                           "run. Program will proceed, but old files may be "
                           "lost.", exc_info=True)
    if os.path.isfile("fortran-compile.log"):
        try:
            os.remove("fortran-compile.log")
        except Exception:
            pass


def organize_workdir(tensor_index, delete_unzipped=False,
                     tensors=True, deltas=True, workdir=Path(),
                     compression_level=2):
    """Reorganize files in workdir into SUPP, OUT, Tensors and Deltas.

    Tensors and Deltas folders are zipped and moved over. All other
    files are copied to appropriate locations in SUPP and OUT.

    Parameters
    ----------
    tensor_index : int
        Which Delta and Tensor files should be considered.
    delete_unzipped : bool, optional
        Whether the original Delta- and Tensor-files should be
        deleted after making the archives. The default is False.
    tensors, deltas : bool, optional
        Whether the Tensor/Delta files contain new information
        and should be saved. The default is True.
    workdir : pathlike, optional
        The path to work folder that contains the files to be
        reorganized. The default is "".
    compression_level : int
        Compression level to be applied for ZIP archives of Tensors and
        Deltas. Default is 2.

    Returns
    -------
    None.
    """
    # outfiles with variable names:
    path = Path(workdir)
    outfiles = set(path / f for f in _OUTFILES)
    for pattern in ("POSCAR_OUT*", "VIBROCC_OUT*", "R_OUT*"):
        outfiles.update(path.glob(pattern))

    _collect_deltas(tensor_index, path)
    _zip_deltas_and_tensors(delete_unzipped, tensors, deltas, path,
                            compression_level)
    _organize_supp_out(path, outfiles)

def _organize_supp_out(path, outfiles):
    # sort SUPP and OUT files:
    for folder in ["SUPP", "OUT"]:
        out_path = path / folder
        try:
            out_path.mkdir(parents=True)
        except FileExistsError:
            pass
        except OSError:
            logger.error(f"Error creating {folder} folder: ", exc_info=True)
            continue

        if folder == "SUPP":
            filelist = set(path / f for f in _SUPP_FILES)
            # move directories original_inputs and compile_logs to SUPP
            directory_list = (path / d for d in _SUPP_DIRS)
            # Also add log files into SUPP: skip calc logs (they go to
            # main dir), and compile logs (they go to compile_logs dir)
            logs_to_supp = (f for f in path.glob("*.log")
                            if (not f.name.startswith(LOG_PREFIX)
                                and "compile" not in f.name))
            filelist.update(logs_to_supp)
        else:
            filelist = outfiles
            directory_list = ()

        for file in filelist:
            if not file.is_file():
                continue
            # copies files into SUPP and OUT directories
            try:
                shutil.copy2(file, out_path / file.name)
            except OSError:
                logger.error(f"Error moving {folder} file {file.name}: ",
                             exc_info=True)

        for _dir in directory_list:
            if not _dir.is_dir():
                continue
            try:
                copytree_exists_ok(_dir, out_path / _dir.name)
            except OSError:
                logger.error(f"Error moving {folder} directory {_dir.name}: ",
                             exc_info=True)


def _zip_deltas_and_tensors(delete_unzipped, tensors, deltas, path,
                            compression_level):
    # If there are unzipped Tensors or Deltas directories, zip them:
    for folder in ["Tensors", "Deltas"]:
        todo = tensors if folder == "Tensors" else deltas
        origin_base = path / folder
        if not origin_base.is_dir():
            continue
        if not todo and not delete_unzipped:
            continue
        rgx = re.compile(rf"{folder}_[0-9]{{3}}")                               # TODO: maybe we want "three or more" digits, i.e., {{3,}}? Or could we use tensor_index?
        for _dir in origin_base.glob("*"):
            if not _dir.is_dir():
                continue
            match = rgx.match(_dir.name)
            if not match or match.span()[1] != len(folder) + 4:                 # TODO: should this 4 be adjusted to the previous TODO? Unclear what it guards
                continue
            delete = delete_unzipped
            if todo:
                logger.info(f"Packing {_dir.name}.zip...")
                _dir_path = Path(_dir)
                move_to_archive = _dir_path.glob('*')
                arch_name = _dir_path.with_suffix(".zip")
                try:
                    with ZipFile(arch_name, 'a', compression=ZIP_DEFLATED,
                        compresslevel=compression_level) as archive:
                        for fname in move_to_archive:
                            archive.write(fname, fname.relative_to(_dir))
                except OSError:
                    logger.error(f"Error packing {_dir.name}.zip file: ",
                                    exc_info=True)
                    delete = False
            if delete:
                try:
                    shutil.rmtree(_dir)
                except OSError:
                    logger.warning(
                        f"Error deleting unzipped {folder} directory. "
                        "This will increase the size of the work folder, "
                        "but not cause any problems.")


def _collect_deltas(tensor_index, path):
    # Clean up deltas
    deltalist = list(path.glob("DEL_*"))
    if len(deltalist) > 0:
        destination = path / "Deltas" / f"Deltas_{tensor_index:03d}"
        try:
            destination.mkdir(parents=True)
        except FileExistsError:
            pass
        except OSError:
            logger.error(f"Failed to create {destination} folder: ",
                         exc_info=True)
        if destination.exists():
            errors = []
            for delta_file in deltalist:
                try:
                    shutil.move(delta_file, destination / delta_file.name)
                except OSError as err:
                    errors.append(err)
            if errors:
                logger.error(f"Error moving Delta files: {errors}")


def move_oldruns(rp, prerun=False):
    """Copy relevant files to a new 'workhistory' subfolder.

    Files are copied from SUPP, OUT and the list in rp.manifest.
    The main log file is excluded.

    Parameters
    ----------
    rp : Rparams
        The run parameters.
    prerun : bool, optional
        If True, then instead of using the manifest, all potentially
        interesting files will be copied, and the new folder will get index 0.
        Then clears out the SUPP and OUT folders and old SUPP and OUT files
        from the main directory.

    Returns
    -------
    None.
    """
    sectionabbrv = {1: "R", 2: "D", 3: "S"}
    try:
        os.makedirs(os.path.join(".", DEFAULT_WORK_HISTORY), exist_ok=True)
    except Exception:
        logger.error(f"Error creating {DEFAULT_WORK_HISTORY} folder: ",
                     exc_info=True)
        raise
    if not prerun:
        rp.manifest.append(DEFAULT_WORK_HISTORY)
    dl = [n for n in os.listdir(DEFAULT_WORK_HISTORY)
          if os.path.isdir(os.path.join(DEFAULT_WORK_HISTORY, n))]
    maxnum = -1
    rgx = re.compile(r't'+'{:03d}'.format(rp.TENSOR_INDEX)+r'.r[0-9]{3}_')             # TODO: would be nicer to use a capture group for the last three digits, used to decide how to number the new folder
    for d in dl:
        m = rgx.match(d)
        if m:
            try:
                i = int(d[6:9])
                if i > maxnum:
                    maxnum = i
            except Exception:
                pass
    if maxnum == -1:
        if prerun:
            num = 0
        else:
            num = 1
    else:
        num = maxnum + 1
    if prerun:
        oldlogfiles = sorted([f for f in os.listdir() if os.path.isfile(f) and
                              f.endswith(".log") and f.startswith(LOG_PREFIX)
                              and f not in rp.manifest])
        if len(oldlogfiles) > 0:
            oldTimeStamp = oldlogfiles[-1][-17:-4]
        else:
            oldTimeStamp = "moved-" + rp.timestamp
        dirname = ("t{:03d}.r{:03d}_previous_".format(rp.TENSOR_INDEX, num)
                   + oldTimeStamp)
    else:
        dirname = "t{:03d}.r{:03d}_".format(rp.TENSOR_INDEX, num)
        for ind in rp.runHistory[len(rp.lastOldruns):]:
            if ind in sectionabbrv:
                dirname += sectionabbrv[ind]
        rp.lastOldruns = rp.runHistory[:]
        dirname += "_" + rp.timestamp

    # make workhistory directory
    work_hist_path = Path(".") / DEFAULT_WORK_HISTORY / dirname
    try:
        os.mkdir(work_hist_path)
    except Exception:
        logger.error(f"Error creating {DEFAULT_WORK_HISTORY} subfolder: ",
                     exc_info=True)
        raise
    if not prerun:
        organize_workdir(rp.TENSOR_INDEX, delete_unzipped=False,
                         tensors=False, deltas=False,
                         compression_level=rp.ZIP_COMPRESSION_LEVEL)
        for dp in rp.domainParams:
            organize_workdir(dp.rp.TENSOR_INDEX, delete_unzipped=False,
                             tensors=False, deltas=False,
                             compression_level=rp.ZIP_COMPRESSION_LEVEL)
    if prerun:
        filelist = [f for f in os.listdir() if os.path.isfile(f) and
                    (f.endswith(".log") or f in _OUTFILES or f in _SUPP_FILES)
                    and f not in rp.manifest and f not in iofiles]
        dirlist = ["SUPP", "OUT"]
    else:
        filelist = [f for f in rp.manifest if os.path.isfile(f) and not
                    (f.startswith(LOG_PREFIX) and f.endswith(".log"))]
        dirlist = [d for d in rp.manifest if os.path.isdir(d) and
                   d not in ["Tensors", "Deltas", DEFAULT_WORK_HISTORY]]
    for f in filelist:
        try:
            if not prerun or f in iofiles:
                shutil.copy2(f, work_hist_path / f)
            else:
                shutil.move(f, work_hist_path / f)
        except Exception:
            logger.warning(f"Error copying {f} to {work_hist_path / f}."
                           " File may get overwritten.")
    for d in dirlist:
        try:
            if not prerun:
                shutil.copytree(d, work_hist_path / d)
            else:
                shutil.move(d, work_hist_path / d)
        except Exception:
            logger.warning(f"Error copying {d} to {work_hist_path / d}."
                           " Files in directory may get overwritten.")
    return


def cleanup(manifest, rp=None):
    """
    Moves files to SUPP and OUT folders, writes manifest, adds a final
    message to the log, then shuts down everything.

    Parameters
    ----------
    manifest : list of str
        The files and directories that should be preserved from the work
        folder.
    rp : Rparams, optional
        The run parameters. If None, it is assumed that the run crashed before
        an Rparams object existed.

    Returns
    -------
    None.

    """

    logger.info("\nStarting cleanup...")
    if rp is None:
        history = []
        to_sort = [{"newTensors": False, "newDeltas": False, "tind": 0,
                    "path": ""}]
        compress_level = 2
    else:
        history = rp.runHistory
        rp.closePdfReportFigs()
        compress_level = rp.ZIP_COMPRESSION_LEVEL
        if not rp.domainParams:
            to_sort = [{"newTensors": ("Tensors" in rp.manifest),
                        "newDeltas": ("Deltas" in rp.manifest),
                        "tind": rp.TENSOR_INDEX, "path": ""}]
        else:
            to_sort = [{"newTensors": False, "newDeltas": False, "tind": 0,
                        "path": ""}]
            for dp in rp.domainParams:
                to_sort.append({"newTensors": ("Tensors" in dp.rp.manifest),
                                "newDeltas": ("Deltas" in dp.rp.manifest),
                                "tind": dp.rp.TENSOR_INDEX,
                                "path": dp.workdir})
    for d in to_sort:
        try:
            organize_workdir(d["tind"], delete_unzipped=True,
                             tensors=d["newTensors"],
                             deltas=d["newDeltas"], workdir=d["path"],
                             compression_level=compress_level)
        except Exception:
            logger.warning("Error sorting files to SUPP/OUT folders: ",
                           exc_info=True)
    # write manifest
    written = []
    try:
        with open("manifest", "w") as wf:
            for fn in manifest:
                if fn not in written:
                    wf.write(fn+"\n")
                    written.append(fn)
        logger.info("Wrote manifest file successfully.")
    except Exception:
        logger.error("Failed to write manifest file.")

    # write final log message
    if rp is None:
        elapsedTimeStr = "unknown"
    else:
        elapsedTimeStr = get_elapsed_time_str(timer() - rp.starttime)
    logger.info("\nFinishing execution at "+time.strftime("%Y-%m-%d %H:%M:%S",
                                                          time.localtime())
                + "\nTotal elapsed time: "+elapsedTimeStr+"\n")
    if len(history) > 0:
        s = ""
        for ind in history:
            s += (str(ind)+" ")
        logger.info("Executed segments: "+s[:-1])
    if rp is not None:
        for t in ["refcalc", "superpos"]:
            if rp.stored_R[t] is not None:
                o = "Final R ({}): {:.4f}".format(t, rp.stored_R[t][0])
                if rp.stored_R[t][1] > 0 and rp.stored_R[t][2] > 0:
                    o += " ({:.4f} / {:.4f})".format(rp.stored_R[t][1],
                                                     rp.stored_R[t][2])
                logger.info(o)
    if rp:
        if rp.checklist:
            logger.info("")
            logger.info("# The following issues should be checked before "
                        "starting again:")
            for s in rp.checklist:
                logger.info("- "+s)
    logger.info("")
    # shut down logger
    while logger.handlers:
        logger.removeHandler(logger.handlers[0])
    logging.shutdown()
    return
